from .. import AD_plugin_primed
from .. import AD_prime_plugin2
from .. import CamMixin_V34
from .. import SingleTrigger_V34
from ophyd import EpicsSignalWithRBV
from ophyd import SingleTrigger
from ophyd.areadetector import ADComponent
from ophyd.areadetector import DetectorBase
from ophyd.areadetector import SimDetectorCam
from ophyd.areadetector.filestore_mixins import FileStoreHDF5IterativeWrite
from ophyd.areadetector.plugins import HDF5Plugin_V34 as HDF5Plugin
from ophyd.areadetector.plugins import ImagePlugin_V34 as ImagePlugin
from packaging import version
import bluesky
import bluesky.plans as bp
import databroker
import pathlib


IOC = "ad:"
IMAGE_DIR = "adsimdet/%Y/%m/%d"
AD_IOC_MOUNT_PATH = pathlib.Path("/tmp")
BLUESKY_MOUNT_PATH = pathlib.Path("/tmp/docker_ioc/iocad/tmp")

# MUST end with a `/`, pathlib will NOT provide it
WRITE_PATH_TEMPLATE = f"{AD_IOC_MOUNT_PATH / IMAGE_DIR}/"
READ_PATH_TEMPLATE = f"{BLUESKY_MOUNT_PATH / IMAGE_DIR}/"


class SimDetectorCam_V34(CamMixin_V34, SimDetectorCam):
    """triggering configuration and AcquireBusy support"""


class MyHDF5Plugin(FileStoreHDF5IterativeWrite, HDF5Plugin):
    """
    Add data acquisition methods to HDF5Plugin.

    * ``stage()`` - prepare device PVs befor data acquisition
    * ``unstage()`` - restore device PVs after data acquisition
    * ``generate_datum()`` - coordinate image storage metadata
    """

    def stage(self):
        self.stage_sigs.move_to_end("capture", last=True)
        super().stage()


class SimDetector_V34(SingleTrigger_V34, DetectorBase):
    """ADSimDetector"""

    cam = ADComponent(SimDetectorCam_V34, "cam1:")
    hdf1 = ADComponent(
        MyHDF5Plugin,
        "HDF1:",
        write_path_template=WRITE_PATH_TEMPLATE,
        read_path_template=READ_PATH_TEMPLATE,
    )
    image = ADComponent(ImagePlugin, "image1:")


def test_cam_mixin_v34_structure():
    adsimdet = SimDetector_V34(IOC, name="adsimdet")
    assert isinstance(adsimdet, DetectorBase)
    assert isinstance(adsimdet.cam, SimDetectorCam_V34)
    assert version.parse(adsimdet.cam._cam_release) >= version.parse("3.4")
    assert adsimdet._acquisition_busy_signal != adsimdet._acquisition_signal
    assert adsimdet.cam.pool_max_buffers is None
    for attr_name in "acquire_busy offset wait_for_plugins".split():
        assert hasattr(adsimdet.cam, attr_name), attr_name
        assert attr_name in adsimdet.cam.component_names, attr_name


def test_the_old_way():
    class MyFixedCam(SimDetectorCam):
        "No need to add unused, new attributes such as offset."
        pool_max_buffers = None

    class OldHDF5Plugin(FileStoreHDF5IterativeWrite, HDF5Plugin):
        pass

    class Phoenix(SingleTrigger, DetectorBase):

        cam = ADComponent(MyFixedCam, "cam1:")
        hdf1 = ADComponent(
            OldHDF5Plugin,
            "HDF1:",
            write_path_template=WRITE_PATH_TEMPLATE,
            read_path_template=READ_PATH_TEMPLATE,
        )
        image = ADComponent(ImagePlugin, "image1:")

    adsimdet = Phoenix(IOC, name="adsimdet")
    adsimdet.wait_for_connection(timeout=15)
    assert adsimdet.connected

    adsimdet.read_attrs.append("hdf1")
    adsimdet.hdf1.create_directory.put(-5)

    NUM_FRAMES = 1
    adsimdet.cam.stage_sigs["acquire_time"] = 0.1
    adsimdet.cam.stage_sigs["image_mode"] = "Multiple"
    adsimdet.cam.stage_sigs["num_images"] = NUM_FRAMES
    adsimdet.hdf1.stage_sigs.move_to_end("capture")

    if not AD_plugin_primed(adsimdet.hdf1):
        AD_prime_plugin2(adsimdet.hdf1)
    assert AD_plugin_primed(adsimdet.hdf1)

    cat = databroker.temp().v2
    RE = bluesky.RunEngine({})
    RE.subscribe(cat.v1.insert)

    # acquire the image (using standard technique)
    uids = RE(bp.count([adsimdet]))
    assert len(uids) == 1

    run = cat.v2[uids[0]]
    assert run is not None

    dataset = run.primary.read()
    assert dataset is not None

    image = dataset["adsimdet_image"]
    assert image is not None
    assert len(image.shape) == 4
    assert image.shape == (1, NUM_FRAMES, 1024, 1024)


def test_ignore_no_WaitForPlugins():
    """This is the old way (pre AD 3.1)."""

    class MyFixedCam(SimDetectorCam):
        "No need to add unused, new attributes such as offset."
        pool_max_buffers = None

    class OldHDF5Plugin(FileStoreHDF5IterativeWrite, HDF5Plugin):
        pass

    class ThisSimDetector(SingleTrigger, DetectorBase):
        """Custom ADSimDetector."""

        cam = ADComponent(MyFixedCam, "cam1:")
        hdf1 = ADComponent(
            OldHDF5Plugin,
            "HDF1:",
            write_path_template=WRITE_PATH_TEMPLATE,
            read_path_template=READ_PATH_TEMPLATE,
        )
        image = ADComponent(ImagePlugin, "image1:")

    adsimdet = ThisSimDetector(IOC, name="adsimdet")
    adsimdet.wait_for_connection(timeout=15)
    assert adsimdet.connected

    adsimdet.read_attrs.append("hdf1")

    adsimdet.hdf1.create_directory.put(-5)

    NUM_FRAMES = 100
    adsimdet.cam.stage_sigs["acquire_period"] = 0.002
    adsimdet.cam.stage_sigs["acquire_time"] = 0.001
    adsimdet.cam.stage_sigs["image_mode"] = "Multiple"
    adsimdet.cam.stage_sigs["num_images"] = NUM_FRAMES

    # ('None', 'N-bit', 'szip', 'zlib', 'Blosc', 'BSLZ4', 'LZ4', 'JPEG')
    # LZ4 and others require 'import hdf5plugin'
    adsimdet.hdf1.stage_sigs["compression"] = "zlib"
    adsimdet.hdf1.stage_sigs["file_template"] = "%s%s_%3.3d.h5"

    # Can't use LazyOpen feature and access run.primary.read() using databroker.
    # Raises event_model.EventModelError: Error instantiating handler class
    #   <class 'area_detector_handlers.handlers.AreaDetectorHDF5Handler'>
    # adsimdet.hdf1.stage_sigs["lazy_open"] = 1

    adsimdet.hdf1.stage_sigs.move_to_end("capture")

    if not AD_plugin_primed(adsimdet.hdf1):
        AD_prime_plugin2(adsimdet.hdf1)

    cat = databroker.temp().v2
    RE = bluesky.RunEngine({})
    RE.subscribe(cat.v1.insert)

    # acquire the image (using standard technique)
    uids = RE(bp.count([adsimdet]))
    assert len(uids) == 1

    run = cat.v2[uids[0]]
    assert run is not None

    dataset = run.primary.read()
    assert dataset is not None

    image = dataset["adsimdet_image"]
    assert image is not None
    assert len(image.shape) == 4
    assert image.shape == (1, NUM_FRAMES, 1024, 1024)


def test_cam_mixin_v34_operation():
    adsimdet = SimDetector_V34(IOC, name="adsimdet")
    adsimdet.wait_for_connection()
    assert adsimdet.connected

    if not AD_plugin_primed(adsimdet.hdf1):
        AD_prime_plugin2(adsimdet.hdf1)
    assert AD_plugin_primed(adsimdet.hdf1)

    adsimdet.read_attrs.append("hdf1")
    adsimdet.hdf1.create_directory.put(-5)

    NUM_FRAMES = 100
    adsimdet.cam.stage_sigs["acquire_period"] = 0.0021
    adsimdet.cam.stage_sigs["acquire_time"] = 0.001
    adsimdet.cam.stage_sigs["image_mode"] = "Multiple"
    adsimdet.cam.stage_sigs["num_images"] = NUM_FRAMES
    adsimdet.cam.stage_sigs["wait_for_plugins"] = "Yes"
    adsimdet.hdf1.stage_sigs["auto_increment"] = "Yes"
    adsimdet.hdf1.stage_sigs["auto_save"] = "Yes"
    adsimdet.hdf1.stage_sigs["blocking_callbacks"] = "No"
    adsimdet.hdf1.stage_sigs["compression"] = "zlib"
    adsimdet.hdf1.stage_sigs["file_template"] = "%s%s_%5.5d.h5"
    adsimdet.hdf1.stage_sigs["num_capture"] = 0  # means: capture ALL frames sent
    adsimdet.hdf1.stage_sigs["queue_size"] = 100

    # Can't use LazyOpen feature and access run.primary.read() using databroker.
    # Raises event_model.EventModelError: Error instantiating handler class
    #   <class 'area_detector_handlers.handlers.AreaDetectorHDF5Handler'>
    # adsimdet.hdf1.stage_sigs["lazy_open"] = 1

    adsimdet.image.stage_sigs["blocking_callbacks"] = "No"

    assert adsimdet.cam.stage_sigs["wait_for_plugins"] == "Yes"
    assert adsimdet.hdf1.stage_sigs["blocking_callbacks"] == "No"

    cat = databroker.temp().v2
    RE = bluesky.RunEngine({})
    RE.subscribe(cat.v1.insert)

    # acquire the image (using standard technique)
    uids = RE(bp.count([adsimdet]))
    assert len(uids) == 1
    assert not adsimdet.cam.is_busy
    assert adsimdet.hdf1.num_captured.get() == NUM_FRAMES

    expected_sigs = {
        adsimdet: {
            "cam.acquire": 0,
            "cam.image_mode": 1,
        },
        adsimdet.cam: {
            "acquire_period": 0.0021,
            "acquire_time": 0.001,
            "image_mode": "Multiple",
            "num_images": NUM_FRAMES,
            "wait_for_plugins": "Yes",
        },
        adsimdet.hdf1: {
            "array_counter": 0,
            "auto_increment": "Yes",
            "auto_save": "Yes",
            "blocking_callbacks": "No",
            "compression": "zlib",
            "create_directory": -3,
            "enable": 1,
            "file_template": "%s%s_%5.5d.h5",
            "file_write_mode": "Stream",
            "num_capture": 0,
            "parent.cam.array_callbacks": 1,
            'queue_size': 100,
            "capture": 1,
        },
        adsimdet.image: {
            "enable": 1,
            "blocking_callbacks": "No",
            "parent.cam.array_callbacks": 1,
        },
    }
    for obj, expected in expected_sigs.items():
        assert obj.stage_sigs == expected, str(obj)

    run = cat.v2[uids[0]]
    assert run is not None

    dataset = run.primary.read()
    assert dataset is not None

    image = dataset["adsimdet_image"]
    assert image is not None
    assert len(image.shape) == 4
    assert image.shape == (1, NUM_FRAMES, 1024, 1024)



"""
test the Eurotherm 2216e device support

Hardware is not available so test with best efforts
"""

from .. import eurotherm_2216e
from ...tests import IOC

PV_PREFIX = f"phony:{IOC}2216e:"


def test_device():
    euro = eurotherm_2216e.Eurotherm2216e(PV_PREFIX, name="controller")
    assert not euro.connected

    assert euro.tolerance.get() == 1
    assert euro.update_target is False
    assert euro.target is None

    cns = """
    readback setpoint
    """.split()
    assert sorted(euro.read_attrs) == sorted(cns)

    cns += """
    sensor
    done tolerance report_dmov_changes
    mode power program_number
    """.split()
    assert sorted(euro.component_names) == sorted(cns)



"""
Ophyd support for the EPICS calcout record

https://wiki-ext.aps.anl.gov/epics/index.php/RRM_3-14_Calcout

Public Structures

.. autosummary::

    ~UserCalcoutDevice
    ~UserCalcoutN
    ~CalcoutRecord
    ~CalcoutRecordChannel
    ~setup_gaussian_calcout
    ~setup_incrementer_calcout
    ~setup_lorentzian_calcout
"""

from collections import OrderedDict
from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as DDC
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import FormattedComponent as FC
from ophyd import Signal

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsRecordFloatFields
from ._common import EpicsSynAppsRecordEnableMixin
from .. import utils as APS_utils


CHANNEL_LETTERS_LIST = "A B C D E F G H I J K L".split()


class CalcoutRecordChannel(Device):
    """
    channel of a calcout record: A-L

    .. index:: Ophyd Device; synApps CalcoutRecordChannel

    .. autosummary::

        ~reset
    """

    input_value = FC(EpicsSignal, "{self.prefix}.{self._ch_letter}", kind="config")
    last_value = FC(EpicsSignalRO, "{self.prefix}.L{self._ch_letter}", kind="config")
    input_pv = FC(EpicsSignal, "{self.prefix}.INP{self._ch_letter}", kind="config")
    input_pv_valid = FC(EpicsSignalRO, "{self.prefix}.IN{self._ch_letter}V", kind="config")

    read_attrs = [
        "input_value",
    ]
    hints = {"fields": read_attrs}

    def __init__(self, prefix, letter, **kwargs):
        self._ch_letter = letter
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.input_pv.put("")
        self.input_value.put(0)


def _channels(channel_list):
    defn = OrderedDict()
    for chan in channel_list:
        defn[chan] = (CalcoutRecordChannel, "", {"letter": chan})
    return defn


class CalcoutRecord(EpicsRecordFloatFields, EpicsRecordDeviceCommonAll):
    """
    EPICS base calcout record support in ophyd

    .. index:: Ophyd Device; synApps CalcoutRecord

    .. autosummary::

        ~reset

    :see: https://wiki-ext.aps.anl.gov/epics/index.php/RRM_3-14_Calcout
    """
    units = Cpt(EpicsSignal, ".EGU", kind="config")
    precision = Cpt(EpicsSignal, ".PREC", kind="config")

    calculated_value = Cpt(EpicsSignal, ".VAL", kind="normal")
    calculation = Cpt(EpicsSignal, ".CALC", kind="config")

    output_pv = Cpt(EpicsSignal, ".OUT", kind="config")
    output_execute_option = Cpt(EpicsSignal, ".OOPT", kind="config")
    output_execution_delay = Cpt(EpicsSignal, ".ODLY", kind="config")
    output_data_option = Cpt(EpicsSignal, ".DOPT", kind="config")
    output_calculation = Cpt(EpicsSignal, ".OCAL", kind="config")
    output_value = Cpt(EpicsSignal, ".OVAL", kind="hinted")
    invalid_output_action = Cpt(EpicsSignal, ".IVOA", kind="config")
    invalid_output_value = Cpt(EpicsSignal, ".IVOV", kind="config")
    event_to_issue = Cpt(EpicsSignal, ".OEVT", kind="config")

    output_pv_status = Cpt(EpicsSignal, ".OUTV", kind="config")
    calculation_valid = Cpt(EpicsSignal, ".CLCV", kind="config")
    output_calculation_valid = Cpt(EpicsSignal, ".OCLV", kind="config")
    output_delay_active = Cpt(EpicsSignal, ".DLYA", kind="config")

    channels = DDC(_channels(CHANNEL_LETTERS_LIST))

    read_attrs = APS_utils.itemizer("channels.%s", CHANNEL_LETTERS_LIST)
    hints = {"fields": read_attrs}

    @property
    def value(self):
        return self.calculated_value.get()

    def reset(self):
        """set all fields to default values"""
        pvname = self.description.pvname.split(".")[0]
        self.scanning_rate.put("Passive")
        self.description.put(pvname)
        self.units.put("")
        self.precision.put("5")

        self.calculation.put("0")
        self.calculated_value.put(0)
        self.output_calculation.put("")
        self.output_value.put(0)

        self.forward_link.put("")
        self.output_pv.put("")
        self.invalid_output_action.put(0)
        self.invalid_output_value.put(0)

        self.output_execution_delay.put(0)
        self.output_execute_option.put(0)
        self.output_data_option.put(0)

        for letter in self.channels.read_attrs:
            channel = getattr(self.channels, letter)
            if isinstance(channel, CalcoutRecordChannel):
                channel.reset()
        self.hints = {"fields": ["channels.%s" % c for c in CHANNEL_LETTERS_LIST]}
        self.read_attrs = ["channels.%s" % c for c in CHANNEL_LETTERS_LIST]


class UserCalcoutN(EpicsSynAppsRecordEnableMixin, CalcoutRecord):
    """Single instance of the userCalcoutN database."""


class UserCalcoutDevice(Device):
    """
    EPICS synApps XXX IOC setup of user calcouts: ``$(P):userCalcOut$(N)``

    .. index:: Ophyd Device; synApps UserCalcoutDevice

    .. autosummary::

        ~reset

    """

    enable = Cpt(EpicsSignal, "userCalcOutEnable", kind="omitted")
    calcout1 = Cpt(UserCalcoutN, "userCalcOut1")
    calcout2 = Cpt(UserCalcoutN, "userCalcOut2")
    calcout3 = Cpt(UserCalcoutN, "userCalcOut3")
    calcout4 = Cpt(UserCalcoutN, "userCalcOut4")
    calcout5 = Cpt(UserCalcoutN, "userCalcOut5")
    calcout6 = Cpt(UserCalcoutN, "userCalcOut6")
    calcout7 = Cpt(UserCalcoutN, "userCalcOut7")
    calcout8 = Cpt(UserCalcoutN, "userCalcOut8")
    calcout9 = Cpt(UserCalcoutN, "userCalcOut9")
    calcout10 = Cpt(UserCalcoutN, "userCalcOut10")

    def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        self.calcout1.reset()
        self.calcout2.reset()
        self.calcout3.reset()
        self.calcout4.reset()
        self.calcout5.reset()
        self.calcout6.reset()
        self.calcout7.reset()
        self.calcout8.reset()
        self.calcout9.reset()
        self.calcout10.reset()
        self.read_attrs = ["calcout%d" % (c + 1) for c in range(10)]
        self.read_attrs.insert(0, "enable")


def _setup_peak_calcout_(
    # fmt: off
    calc, desc, calcout, ref_signal,
    center=0, width=1, scale=1, noise=0.05
    # fmt: on
):
    """
    internal: setup that is common to both Gaussian and Lorentzian calcouts

    PARAMETERS

    calcout
        *object* :
        instance of :class:`CalcoutRecord`

    ref_signal
        *object* :
        instance of :class:`EpicsSignal` used as ``A``

    center
        *float* :
        EPICS record field ``B``,
        default = 0

    width
        *float* :
        EPICS record field ``C``,
        default = 1

    scale
        *float* :
        EPICS record field ``D``,
        default = 1

    noise
        *float* :
        EPICS record field ``E``,
        default = 0.05
    """

    # to add a noisy background will need another calc
    if not isinstance(calcout, CalcoutRecord):
        raise TypeError(f"expected CalcoutRecord instance, received {type(calcout)}")
    if not isinstance(ref_signal, Signal):
        raise TypeError(f"expected Signal instance, received {type(ref_signal)}")
    if width <= 0:
        raise ValueError(f"width must be positive, received {width}")
    if not (0.0 <= noise <= 1.0):
        raise ValueError(f"noise must be between 0 and 1, received {noise}")

    calcout.reset()
    calcout.scanning_rate.put("Passive")
    calcout.description.put(desc)
    calcout.channels.A.input_pv.put(ref_signal.pvname)
    calcout.channels.B.input_value.put(center)
    calcout.channels.C.input_value.put(width)
    calcout.channels.D.input_value.put(scale)
    calcout.channels.E.input_value.put(noise)
    calcout.calculation.put(calc)
    calcout.scanning_rate.put(".1 second")

    calcout.read_attrs = [
        "input_value",
    ]
    calcout.hints = {"fields": calcout.read_attrs}


def setup_gaussian_calcout(calcout, ref_signal, center=0, width=1, scale=1, noise=0.05):
    """
    setup calcout for noisy Gaussian

    calculation::

        D*(0.95+E*RNDM)/exp(((A-B)/C)^2)

    PARAMETERS

    calcout
        *object* :
        instance of :class:`CalcoutRecord`

    ref_signal
        *object* :
        instance of :class:`EpicsSignal` used as ``A``

    center
        *float* :
        EPICS record field ``B``,
        default = 0

    width
        *float* :
        EPICS record field ``C``,
        default = 1

    scale
        *float* :
        EPICS record field ``D``,
        default = 1

    noise
        *float* :
        EPICS record field ``E``,
        default = 0.05
    """
    _setup_peak_calcout_(
        "D*(0.95+E*RNDM)/exp(((A-b)/c)^2)",
        "noisy Gaussian curve",
        calcout,
        ref_signal,
        center=center,
        width=width,
        scale=scale,
        noise=noise,
    )


def setup_lorentzian_calcout(
    calcout, ref_signal, center=0, width=1, scale=1, noise=0.05
):  # lgtm [py/similar-function]
    """
    setup calcout record for noisy Lorentzian

    calculation::

        D*(0.95+E*RNDM)/(1+((A-B)/C)^2)

    PARAMETERS

    calcout
        *object* :
        instance of :class:`CalcoutRecord`

    ref_signal
        *object* :
        instance of :class:`EpicsSignal` used as ``A``

    center
        *float* :
        EPICS record field ``B``,
        default = 0

    width
        *float* :
        EPICS record field ``C``,
        default = 1

    scale
        *float* :
        EPICS record field ``D``,
        default = 1

    noise
        *float* :
        EPICS record field ``E``,
        default = 0.05
    """
    _setup_peak_calcout_(
        "D*(0.95+E*RNDM)/(1+((A-B)/C)^2)",
        "noisy Lorentzian curve",
        calcout,
        ref_signal,
        center=center,
        width=width,
        scale=scale,
        noise=noise,
    )


def setup_incrementer_calcout(calcout, scan=None, limit=100000):
    """
    setup calcout record as an incrementer

    PARAMETERS

    calcout
        *object* :
        instance of :class:`CalcoutRecord`

    scan
        *text* or *int* or ``None`` :
        any of the EPICS record ``.SCAN`` values,
        or the index number of the value,
        set to default if ``None``,
        default: ``.1 second``

    limit
        *int* or ``None`` :
        set the incrementer back to zero
        when this number is reached (or passed),
        default: 100000

    """
    # consider a noisy background, as well (needs a couple calcs)
    scan = scan or ".1 second"
    calcout.reset()
    calcout.scanning_rate.put("Passive")
    calcout.description.put("incrementer")
    pvname = calcout.calculated_value.pvname.split(".")[0]
    calcout.channels.A.input_pv.put(pvname)
    calcout.channels.B.input_value.put(limit)
    calcout.calculation.put("(A+1) % B")
    calcout.scanning_rate.put(scan)

    calcout.hints = {"fields": ["input_value", ]}
    calcout.read_attrs = [
        "input_value",
    ]

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
test the Eurotherm 2216e device support

Hardware is not available so test with best efforts
"""

from ..lakeshore_controllers import LakeShore336Device, LakeShore340Device

IOC = "gp:"
PV_PREFIX = f"phony:{IOC}lakeshore:"


def test_lakeshore_336():
    t336 = LakeShore336Device(PV_PREFIX, name="lakeshore336")
    assert not t336.connected

    # Signal components
    assert t336.loop1.target.get() == "None"
    assert t336.loop2.target.get() == "None"

    assert t336.loop1.tolerance.get() == 0.1
    assert t336.loop2.tolerance.get() == 0.1

    assert t336.loop1.settle_time == 0
    assert t336.loop2.settle_time == 0

    # Read attrs
    cns = """
    loop1 loop1.readback loop1.setpoint loop1.heater loop2 loop2.readback
    loop2.setpoint loop2.heater loop3 loop3.readback loop4 loop4.readback
    """.split()
    assert sorted(t336.read_attrs) == sorted(cns)

    # Components
    cns = """
    loop1 loop2 loop3 loop4 scanning_rate process_record read_all serial
    """.split()
    assert sorted(t336.component_names) == sorted(cns)

    cns = """
    readback setpoint done tolerance report_dmov_changes target units heater
    heater_range pid_P pid_I pid_D ramp_rate ramp_on loop_name control manual
    mode
    """.split()
    assert sorted(t336.loop1.component_names) == sorted(cns)

    cns = """
    readback setpoint done tolerance report_dmov_changes target units heater
    heater_range pid_P pid_I pid_D ramp_rate ramp_on loop_name control manual
    mode
    """.split()
    assert sorted(t336.loop2.component_names) == sorted(cns)

    cns = """
    readback units loop_name
    """.split()
    assert sorted(t336.loop3.component_names) == sorted(cns)

    cns = """
    readback units loop_name
    """.split()
    assert sorted(t336.loop3.component_names) == sorted(cns)


def test_lakeshore_340():
    t340 = LakeShore340Device(PV_PREFIX, name="lakeshore340")
    assert not t340.connected

    # Signal components
    assert t340.control.target.get() == "None"
    assert t340.sample.target.get() == "None"

    assert t340.control.tolerance.get() == 0.1
    assert t340.sample.tolerance.get() == 0.1

    assert t340.control.settle_time == 0
    assert t340.sample.settle_time == 0

    # Read attrs
    cns = """
    control control.readback control.setpoint sample sample.readback
    sample.setpoint heater heater_range
    """.split()
    assert sorted(t340.read_attrs) == sorted(cns)

    # Components
    cns = """
    control sample heater heater_range read_pid scanning_rate process_record
    serial
    """.split()
    assert sorted(t340.component_names) == sorted(cns)

    cns = """
    readback setpoint done tolerance report_dmov_changes target units pid_P
    pid_I pid_D ramp_rate ramp_on sensor
    """.split()
    assert sorted(t340.control.component_names) == sorted(cns)

    cns = """
    readback setpoint done tolerance report_dmov_changes target units pid_P
    pid_I pid_D ramp_rate ramp_on sensor
    """.split()
    assert sorted(t340.sample.component_names) == sorted(cns)



from .alignment import lineup
from .alignment import tune_axes
from .alignment import TuneAxis
from .alignment import TuneResults

from .command_list import command_list_as_table
from .command_list import CommandFileReadError
from .command_list import execute_command_list
from .command_list import get_command_list
from .command_list import parse_Excel_command_file
from .command_list import parse_text_command_file
from .command_list import register_command_handler
from .command_list import run_command_file
from .command_list import summarize_command_file

from .doc_run import addDeviceDataAsStream
from .doc_run import documentation_run

from .input_plan import request_input

from .nscan_support import nscan

from .run_blocking_function_plan import run_blocking_function

from .sscan_support import sscan_1D

from .stage_sigs_support import restorable_stage_sigs
from .stage_sigs_support import stage_sigs_wrapper

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Test code where users control the names of the image files using EPICS.
"""

from .. import AD_EpicsFileNameMixin
from .. import AD_EpicsFileNameHDF5Plugin
from .. import AD_EpicsFileNameJPEGPlugin
from .. import AD_EpicsFileNameTIFFPlugin
from .. import AD_full_file_name_local
from .. import AD_plugin_primed
from .. import AD_prime_plugin2
from .. import CamMixin_V34 as CamMixin
from .. import SingleTrigger_V34 as SingleTrigger
from ...tests import MASTER_TIMEOUT
from ophyd.areadetector import ADComponent
from ophyd.areadetector import DetectorBase
from ophyd.areadetector import SimDetectorCam
from ophyd.areadetector.plugins import ImagePlugin_V34 as ImagePlugin
from ophyd.areadetector.plugins import PvaPlugin_V34 as PvaPlugin
from ophyd.signal import EpicsSignalBase
import bluesky
import bluesky.plan_stubs as bps
import bluesky.plans as bp
import datetime
import pathlib
import pytest
import time


IOC = "ad:"
IMAGE_DIR = "adsimdet/%Y/%m/%d"
AD_IOC_MOUNT_PATH = pathlib.Path("/tmp")
BLUESKY_MOUNT_PATH = pathlib.Path("/tmp/docker_ioc/iocad/tmp")

# MUST end with a `/`, pathlib will NOT provide it
WRITE_PATH_TEMPLATE = f"{AD_IOC_MOUNT_PATH / IMAGE_DIR}/"
READ_PATH_TEMPLATE = f"{BLUESKY_MOUNT_PATH / IMAGE_DIR}/"


# set default timeout for all EpicsSignal connections & communications
try:
    EpicsSignalBase.set_defaults(
        auto_monitor=True,
        timeout=MASTER_TIMEOUT,
        write_timeout=MASTER_TIMEOUT,
        connection_timeout=MASTER_TIMEOUT,
    )
except RuntimeError:
    pass  # ignore if some EPICS object already created


class MySimDetectorCam(CamMixin, SimDetectorCam):
    """triggering configuration and AcquireBusy support"""


class MySimDetector(SingleTrigger, DetectorBase):
    """ADSimDetector"""

    cam = ADComponent(MySimDetectorCam, "cam1:")
    hdf1 = ADComponent(
        AD_EpicsFileNameHDF5Plugin,
        "HDF1:",
        write_path_template=WRITE_PATH_TEMPLATE,
        read_path_template=READ_PATH_TEMPLATE,
    )
    image = ADComponent(ImagePlugin, "image1:")
    jpeg1 = ADComponent(
        AD_EpicsFileNameJPEGPlugin,
        "JPEG1:",
        write_path_template=WRITE_PATH_TEMPLATE,
        read_path_template=READ_PATH_TEMPLATE,
    )
    pva = ADComponent(PvaPlugin, "Pva1:")
    tiff1 = ADComponent(
        AD_EpicsFileNameTIFFPlugin,
        "TIFF1:",
        write_path_template=WRITE_PATH_TEMPLATE,
        read_path_template=READ_PATH_TEMPLATE,
    )


def test_AD_EpicsFileNameMixin():
    # TODO: improve this test per issue #744
    assert AD_EpicsFileNameMixin is not None


@pytest.fixture(scope="function")
def adsimdet():
    "EPICS ADSimDetector."
    adsimdet = MySimDetector(IOC, name="adsimdet")
    adsimdet.stage_sigs["cam.wait_for_plugins"] = "Yes"
    adsimdet.wait_for_connection(timeout=15)
    for plugin_name in "hdf1 jpeg1 tiff1".split():
        adsimdet.read_attrs.append(plugin_name)
        plugin = getattr(adsimdet, plugin_name)
        if not AD_plugin_primed(plugin):
            AD_prime_plugin2(plugin)

        # these settings are the caller's responsibility
        plugin.file_name.put("")  # control this in tests
        plugin.create_directory.put(-5)
        plugin.file_path.put(f"{plugin.write_path_template}/")
        ext = dict(hdf1="h5", jpeg1="jpg", tiff1="tif")[plugin_name]
        plugin.file_template.put(f"%s%s_%4.4d.{ext}")
        plugin.auto_increment.put("Yes")
        plugin.auto_save.put("Yes")
        if plugin_name == "hdf1":
            plugin.compression.put("zlib")
            plugin.zlevel.put(6)
        elif plugin_name == "jpeg":
            plugin.jpeg_quality.put(50)
    yield adsimdet


@pytest.fixture(scope="function")
def fname():
    "File (base) name."
    dt = datetime.datetime.now()
    fname = f"test-image-{dt.minute:02d}-{dt.second:02d}"
    yield fname


@pytest.mark.parametrize(
    "plugin_name, plugin_class",
    [
        ["hdf1", AD_EpicsFileNameHDF5Plugin],
        ["jpeg1", AD_EpicsFileNameJPEGPlugin],
        ["tiff1", AD_EpicsFileNameTIFFPlugin],
    ],
)
def test_stage(plugin_name, plugin_class, adsimdet, fname):
    assert isinstance(plugin_name, str)

    plugin = getattr(adsimdet, plugin_name)
    assert isinstance(plugin, plugin_class)

    plugin.file_name.put(fname)
    time.sleep(0.2)
    assert fname == plugin.file_name.get(), "file name was not put()"

    plugin.stage()
    assert fname == plugin.file_name.get(), "file name was changed by stage()"

    plugin.unstage()
    assert fname == plugin.file_name.get(), "file name was changed by unstage()"


@pytest.mark.parametrize(
    "plugin_name, plugin_class",
    [
        ["hdf1", AD_EpicsFileNameHDF5Plugin],
        ["jpeg1", AD_EpicsFileNameJPEGPlugin],
        ["tiff1", AD_EpicsFileNameTIFFPlugin],
    ],
)
def test_bp_count_custom_name(plugin_name, plugin_class, adsimdet, fname):
    assert isinstance(plugin_name, str)

    plugin = getattr(adsimdet, plugin_name)
    assert isinstance(plugin, plugin_class)

    def prepare_count(file_path, file_name, n_images):
        # fmt: off
        yield from bps.mv(
            plugin.array_counter, 0,
            plugin.auto_increment, "Yes",
            plugin.auto_save, "Yes",
            plugin.create_directory, -5,
            plugin.dropped_arrays, 0,
            plugin.file_name, file_name,
            plugin.file_path, file_path,
            plugin.num_capture, n_images,
            adsimdet.cam.num_images, n_images,
            adsimdet.cam.acquire_time, 0.001,
            adsimdet.cam.acquire_period, 0.002,
            adsimdet.cam.image_mode, "Single",
            # plugin.file_template  # pre-configured
        )
        # fmt: on

    RE = bluesky.RunEngine({})
    NUM_FRAMES = 1
    RE(prepare_count(f"{plugin.write_path_template}/", fname, NUM_FRAMES))
    assert plugin.auto_increment.get() in (1, "Yes")
    assert plugin.auto_save.get() in (1, "Yes")
    assert plugin.create_directory.get() == -5
    assert plugin.file_name.get() == fname
    assert plugin.file_path.get() == f"{plugin.write_path_template}/"
    assert plugin.file_template.get() not in ("", "/")
    assert plugin.num_capture.get() == NUM_FRAMES
    assert adsimdet.cam.num_images.get() == NUM_FRAMES
    assert adsimdet.cam.acquire_time.get() == 0.001
    assert adsimdet.cam.acquire_period.get() == 0.002
    assert adsimdet.cam.image_mode.get() in (0, "Single")

    # remember this for testing later
    next_file_number = plugin.file_number.get()

    # acquire the image (using standard technique)
    uids = RE(bp.count([adsimdet]))
    assert len(uids) == 1
    assert plugin.file_number.get() > next_file_number
    assert not adsimdet.cam.is_busy

    assert plugin.dropped_arrays.get() == 0
    assert plugin.array_counter.get() == NUM_FRAMES
    assert plugin.num_captured.get() == NUM_FRAMES

    # diagnostic per issue #696
    ffname = plugin.full_file_name.get(use_monitor=False).strip()
    if ffname in ("", "/"):
        # second chance, race condition?  don't trust the CA monitor cache
        ffname = plugin.full_file_name.get(use_monitor=False).strip()
    assert ffname not in ("", "/"), str(ffname)

    # local file-system image file name
    lfname = AD_full_file_name_local(plugin)
    assert lfname is not None
    assert lfname.exists(), lfname
    assert isinstance(lfname, pathlib.Path)
    assert fname in str(lfname), f"{lfname=}  {fname=}"


def test_full_file_name_local(adsimdet, fname):
    plugin = adsimdet.hdf1
    plugin.file_name.put(fname)
    plugin.file_path.put(f"{plugin.write_path_template}/")
    time.sleep(0.1)

    adsimdet.stage()  # creates the HDF5 file
    adsimdet.unstage()

    lfname = AD_full_file_name_local(plugin)
    assert lfname is not None, str(plugin)
    assert lfname.exists(), lfname
    assert isinstance(lfname, pathlib.Path)
    assert str(lfname).find(fname) > 0


def test_no_file_path(adsimdet, fname):
    adsimdet.hdf1.file_name.put(fname)
    adsimdet.hdf1.file_path.put("/no-such-path-exists")
    time.sleep(0.1)
    assert adsimdet.hdf1.file_path_exists.get() == 0

    adsimdet.hdf1.file_path.put("")
    time.sleep(0.1)
    assert adsimdet.hdf1.file_path_exists.get() == 0

    with pytest.raises(OSError) as exc:
        adsimdet.stage()
    assert str(exc.value).endswith(" does not exist on IOC.")
    adsimdet.unstage()


def test_file_numbering(adsimdet, fname):
    plugin = adsimdet.hdf1
    plugin.file_name.put(fname)
    plugin.file_path.put(f"{plugin.write_path_template}/")
    time.sleep(0.1)
    assert fname == plugin.file_name.get(), "file name was not put()"

    assert "file_template" not in plugin.stage_sigs

    # prepare for file name with file number
    file_number = plugin.file_number.get()
    plugin.file_template.put("%s%s_%4.4d.h5")
    adsimdet.stage()
    adsimdet.unstage()
    full_file_name = plugin.full_file_name.get()
    assert isinstance(full_file_name, str)
    assert len(full_file_name) > 0
    assert full_file_name.find(fname) > 0, str(full_file_name)
    assert full_file_name.endswith(
        f"{fname}_{file_number:04d}.h5"
    ), f"{full_file_name=} {fname}_{file_number:04d}.h5"

    # prepare for file name with no file number
    file_number = plugin.file_number.get()
    plugin.file_template.put("%s%s.h5")
    time.sleep(0.1)
    adsimdet.stage()
    adsimdet.unstage()
    full_file_name = plugin.full_file_name.get()
    assert isinstance(full_file_name, str)
    assert full_file_name.find(fname) > 0, f"{fname=} {full_file_name=}"
    assert full_file_name.endswith(f"{fname}.h5"), f"{full_file_name=} {fname}.h5"



"""
Plot Support
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~select_mpl_figure
   ~select_live_plot
   ~trim_plot_lines
   ~trim_plot_by_name
"""

import logging


logger = logging.getLogger(__name__)


def select_mpl_figure(x, y):
    """
    Get the MatPlotLib Figure window for y vs x.

    PARAMETERS

    x
        *object*:
        X axis object (an ``ophyd.Signal``)
    y
        ophyd object:
        X axis object (an ``ophyd.Signal``)

    RETURNS

    object or ``None``:
        Instance of ``matplotlib.pyplot.Figure()``
    """
    import matplotlib.pyplot as plt

    figure_name = f"{y.name} vs {x.name}"
    if figure_name in plt.get_figlabels():
        return plt.figure(figure_name)


def select_live_plot(bec, signal):
    """
    Get the *first* live plot that matches ``signal``.

    PARAMETERS

    bec
        *object*:
        instance of ``bluesky.callbacks.best_effort.BestEffortCallback``
    signal
        *object*:
        The Y axis object (an ``ophyd.Signal``)

    RETURNS

    *object*:
        Instance of ``bluesky.callbacks.best_effort.LivePlotPlusPeaks()``
        or ``None``
    """
    for live_plot_dict in bec._live_plots.values():
        live_plot = live_plot_dict.get(signal.name)
        if live_plot is not None:
            return live_plot


def trim_plot_lines(bec, n, x, y):
    """
    Find the plot with axes x and y and replot with at most the last *n* lines.

    Note: :func:`trim_plot_lines` is not a bluesky plan.  Call it as
    normal Python function.

    EXAMPLE::

        trim_plot_lines(bec, 1, m1, noisy)

    PARAMETERS

    bec
        *object* :
        instance of BestEffortCallback

    n
        *int* :
        number of plots to keep

    x
        *object* :
        instance of ophyd.Signal (or subclass),
        independent (x) axis

    y
        *object* :
        instance of ophyd.Signal (or subclass),
        dependent (y) axis

    (new in release 1.3.5)
    """
    liveplot = select_live_plot(bec, y)
    if liveplot is None:
        logger.debug("no live plot found with signal '%s'", y.name)
        return

    fig = select_mpl_figure(x, y)
    if fig is None:
        logger.debug("no figure found with '%s vs %s'", y.name, x.name)
        return
    if len(fig.axes) == 0:
        logger.debug("no plots on figure: '%s vs %s'", y.name, x.name)
        return

    ax = fig.axes[0]
    while len(ax.lines) > n:
        try:
            ax.lines[0].remove()
        except ValueError as exc:
            if not str(exc).endswith("x not in list"):
                # fmt: off
                logger.warning(
                    "%s vs %s: mpl remove() error: %s",
                    y.name, x.name, str(exc),
                )
                # fmt: on
    ax.legend()
    liveplot.update_plot()
    logger.debug("trim complete")


def trim_plot_by_name(n=3, plots=None):
    """
    Find the plot(s) by name and replot with at most the last *n* lines.

    Note: this is not a bluesky plan.  Call it as normal Python function.

    It is recommended to call :func:`~trim_plot_by_name()` *before* the
    scan(s) that generate plots.  Plots are generated from a RunEngine
    callback, executed *after* the scan completes.

    PARAMETERS

    n
        *int* :
        number of plots to keep

    plots
        *str*, [*str*], or *None* :
        name(s) of plot windows to trim
        (default: all plot windows)

    EXAMPLES::

        trim_plot_by_name()   # default of n=3, apply to all plots
        trim_plot_by_name(5)  # change from default of n=3
        trim_plot_by_name(5, "noisy_det vs motor")  # just this plot
        trim_plot_by_name(
            5,
            ["noisy_det vs motor", "det noisy_det vs motor"]]
        )

    EXAMPLE::

        # use simulators from ophyd
        from bluesky import plans as bp
        from bluesky import plan_stubs as bps
        from ophyd.sim import *

        snooze = 0.25

        def scan_set():
            trim_plot_by_name()
            yield from bp.scan([noisy_det], motor, -1, 1, 5)
            yield from bp.scan([noisy_det, det], motor, -2, 1, motor2, 3, 1, 6)
            yield from bps.sleep(snooze)

        # repeat the_scans 15 times
        uids = RE(bps.repeat(scan_set, 15))

    (new in release 1.3.5)
    """
    import matplotlib.pyplot as plt

    if isinstance(plots, str):
        plots = [plots]

    for fig_name in plt.get_figlabels():
        if plots is None or fig_name in plots:
            fig = plt.figure(fig_name)
            for ax in fig.axes:
                while len(ax.lines) > n:
                    ax.lines[0].remove()
                # update the plot legend
                ax.legend()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Test the command list support.
"""

import pathlib
import pyRestTable
import pytest

from .. import CommandFileReadError
from .. import get_command_list
from .. import command_list_as_table

DATA_PATH = pathlib.Path(__file__).parent


@pytest.fixture(scope="function")
def text_command_file():
    text_command_file = (DATA_PATH / "actions.txt")
    return text_command_file


@pytest.fixture(scope="function")
def xl_command_file():
    xl_command_file = (DATA_PATH / "actions.xlsx")
    return xl_command_file


@pytest.fixture(scope="function")
def xl_file():
    xl_file = (DATA_PATH / "demo3.xlsx")
    return xl_file


def test_testfile_exists(text_command_file, xl_command_file, xl_file):
    assert text_command_file.exists()
    assert xl_command_file.exists()
    assert xl_file.exists()


def compare_tables_as_str(expected, received):
    e_lines = expected.strip().splitlines()
    r_lines = received.strip().splitlines()
    assert len(e_lines) == len(r_lines)
    for expect, got in zip(e_lines, r_lines):
        assert expect.strip() == got.strip()


def test_ExcelFile_commands(xl_file):
    # just a spreadsheet for testing (early version of a command file)
    commands = get_command_list(xl_file)
    assert len(commands) == 9  # rows

    table = command_list_as_table(commands, show_raw=False)
    assert isinstance(table, pyRestTable.Table)
    received = str(table).strip()
    expected = """
    ====== ====== =========================================
    line # action parameters
    ====== ====== =========================================
    1      row1   91, 26, 85, None, blank, 8
    2      row2   9, 39, 29, 85, sample, 60
    3      row3   54, None, 38, 3, blank, 76
    4      row4   71, 36, 95, 83, foil, 12
    5      row5   55, 75, 59, 84, DNA, 34
    6      row6   18, 49, 31, 34, lecithin, 47
    7      row7   37, None, None, None, a big mix  of stuff
    8      row8   37, 80, 79, 45, salt water, 36
    9      row9   72, 98, 67, 89, surprises, 49
    ====== ====== =========================================
    """.strip()
    compare_tables_as_str(expected, received)


def test_ExcelCommandList(xl_command_file):
    commands = get_command_list(xl_command_file)
    assert len(commands) == 7
    table = command_list_as_table(commands, show_raw=False)
    received = str(table).strip()
    expected = """
    ====== ============ ===========================
    line # action       parameters
    ====== ============ ===========================
    1      mono_shutter open
    2      USAXSscan    45.07, 98.3, 0, Water Blank
    3      saxsExp      45.07, 98.3, 0, Water Blank
    4      waxwsExp     45.07, 98.3, 0, Water Blank
    5      USAXSscan    12, 12, 1.2, plastic
    6      USAXSscan    12, 37, 0.1, Al foil
    7      mono_shutter close
    ====== ============ ===========================
    """.strip()
    compare_tables_as_str(expected, received)


def test_TextCommandList(text_command_file):
    commands = get_command_list(text_command_file)
    assert len(commands) == 5
    table = command_list_as_table(commands, show_raw=False)
    received = str(table).strip()
    expected = """
    ====== ============ ========================
    line # action       parameters
    ====== ============ ========================
    5      sample_slits 0, 0, 0.4, 1.2
    7      preusaxstune
    10     FlyScan      0, 0, 0, blank
    11     FlyScan      5, 2, 0, empty container
    12     SAXS         0, 0, 0, blank
    ====== ============ ========================
    """.strip()
    compare_tables_as_str(expected, received)


def test_TextCommandListRaw(text_command_file):
    commands = get_command_list(text_command_file)
    assert len(commands) == 5
    table = command_list_as_table(commands, show_raw=True)
    received = str(table).strip()
    expected = """
    ====== ============ ======================== =====================================
    line # action       parameters               raw input
    ====== ============ ======================== =====================================
    5      sample_slits 0, 0, 0.4, 1.2           sample_slits 0 0 0.4 1.2
    7      preusaxstune                          preusaxstune
    10     FlyScan      0, 0, 0, blank           FlyScan 0   0   0   blank
    11     FlyScan      5, 2, 0, empty container FlyScan 5   2   0   "empty container"
    12     SAXS         0, 0, 0, blank           SAXS 0 0 0 blank
    ====== ============ ======================== =====================================
    """.strip()
    compare_tables_as_str(expected, received)


@pytest.mark.parametrize(
    "error, expected, item",
    [
        (IOError, "file not found: ", (DATA_PATH / "none such")),
        (CommandFileReadError, "could not read ", __file__),
    ],
)
def test_known_exceptions(error, expected, item):
    with pytest.raises(error) as exc:
        get_command_list(item)
    assert str(exc.value).startswith(expected)



"""
Ophyd support for the EPICS asyn record

Public Structures

.. autosummary::

    ~AsynRecord

:see: https://github.com/epics-modules/asyn
"""

from ophyd.device import Component
from ophyd import EpicsSignal, EpicsSignalRO
from ._common import EpicsRecordDeviceCommonAll


class AsynRecord(EpicsRecordDeviceCommonAll):
    """
    EPICS asyn record support in ophyd

    .. index:: Ophyd Device; synApps AsynRecord

    :see: https://epics.anl.gov/modules/soft/asyn/R4-36/asynRecord.html
    :see: https://github.com/epics-modules/asyn
    :see: https://epics.anl.gov/modules/soft/asyn/
    """

    ascii_output = Component(EpicsSignal, ".AOUT", kind="hinted")
    binary_output = Component(EpicsSignal, ".BOUT", kind="normal")
    end_of_message_reason = Component(EpicsSignalRO, ".EOMR", kind="config")
    input_format = Component(EpicsSignalRO, ".IFMT", kind="config")
    input_maxlength = Component(EpicsSignal, ".IMAX", kind="config")
    interface = Component(EpicsSignal, ".IFACE", kind="config")
    number_bytes_actually_read = Component(EpicsSignalRO, ".NRRD", kind="normal")
    number_bytes_actually_written = Component(EpicsSignalRO, ".NAWT", kind="normal")
    number_bytes_to_read = Component(EpicsSignal, ".NORD", kind="config")
    number_bytes_to_write = Component(EpicsSignal, ".NOWT", kind="config")
    octet_is_valid = Component(EpicsSignalRO, ".OCTETIV", kind="normal")
    output_format = Component(EpicsSignalRO, ".OFMT", kind="config")
    output_maxlength = Component(EpicsSignal, ".OMAX", kind="config")
    terminator_input = Component(EpicsSignal, ".IEOS", kind="config")
    terminator_output = Component(EpicsSignal, ".OEOS", kind="config")
    timeout = Component(EpicsSignal, ".TMOT", kind="config")
    transaction_mode = Component(EpicsSignal, ".TMOD", kind="config")
    translated_input = Component(EpicsSignal, ".TINP", kind="config")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Unit testing of plans.
"""

from bluesky.simulators import summarize_plan
from ophyd.signal import EpicsSignalBase

import ophyd.sim
import pathlib
import pytest
import sys

from .. import addDeviceDataAsStream
from .. import execute_command_list
from .. import get_command_list
from .. import register_command_handler
from .. import run_command_file
from ...tests import MASTER_TIMEOUT


# set default timeout for all EpicsSignal connections & communications
try:
    EpicsSignalBase.set_defaults(
        auto_monitor=True,
        timeout=MASTER_TIMEOUT,
        write_timeout=MASTER_TIMEOUT,
        connection_timeout=MASTER_TIMEOUT,
    )
except RuntimeError:
    pass  # ignore if some EPICS object already created


DATA_PATH = pathlib.Path(__file__).parent


def test_myoutput(capsys):  # or use "capfd" for fd-level
    print("hello")
    sys.stderr.write("world\n")
    captured = capsys.readouterr()
    assert captured.out == "hello\n"
    assert captured.err == "world\n"
    print("next")
    captured = capsys.readouterr()
    assert captured.out == "next\n"


@pytest.mark.parametrize(
    "objects, name, expected",
    [
        ([ophyd.sim.motor1], "test-device", "  Read ['motor1']\n"),
        (
            [[ophyd.sim.motor2, ophyd.sim.motor3]],
            "test-device-list",
            "  Read ['motor2', 'motor3']\n",
        ),
    ],
)
def test_addDeviceDataAsStream(objects, name, expected, capsys):
    summarize_plan(addDeviceDataAsStream(*objects, name))

    captured = capsys.readouterr()
    assert captured.out == expected


def test_register_action_handler():
    from ..command_list import COMMAND_LIST_REGISTRY

    assert COMMAND_LIST_REGISTRY.command == execute_command_list
    assert summarize_plan != execute_command_list

    register_command_handler(summarize_plan)
    assert COMMAND_LIST_REGISTRY.command == summarize_plan
    assert COMMAND_LIST_REGISTRY.command != execute_command_list

    register_command_handler()
    assert COMMAND_LIST_REGISTRY.command == execute_command_list


def test_get_command_list(capsys):
    filename = (DATA_PATH / "actions.txt")
    assert filename.exists()

    expected = [
        "sample_slits 0 0 0.4 1.2",
        "preusaxstune",
        "FlyScan 0   0   0   blank",
        'FlyScan 5   2   0   "empty container"',
        "SAXS 0 0 0 blank",
    ]

    commands = get_command_list(filename)
    assert len(commands) == len(expected)
    for r, e in zip(commands, expected):
        assert r[-1] == e


def test_run_command_file_text(capsys):
    filename = (DATA_PATH / "actions.txt")
    assert filename.exists()

    expected = [
        f"Command file: {filename}",
        "====== ============ ========================",
        "line # action       parameters              ",
        "====== ============ ========================",
        "5      sample_slits 0, 0, 0.4, 1.2          ",
        "7      preusaxstune                         ",
        "10     FlyScan      0, 0, 0, blank          ",
        "11     FlyScan      5, 2, 0, empty container",
        "12     SAXS         0, 0, 0, blank          ",
        "====== ============ ========================",
        "",
        "file line 5: sample_slits 0 0 0.4 1.2",
        "no handling for line 5: sample_slits 0 0 0.4 1.2",
        "file line 7: preusaxstune",
        "no handling for line 7: preusaxstune",
        "file line 10: FlyScan 0   0   0   blank",
        "no handling for line 10: FlyScan 0   0   0   blank",
        'file line 11: FlyScan 5   2   0   "empty container"',
        'no handling for line 11: FlyScan 5   2   0   "empty container"',
        "file line 12: SAXS 0 0 0 blank",
        "no handling for line 12: SAXS 0 0 0 blank",
    ]

    summarize_plan(run_command_file(filename))

    captured = capsys.readouterr()
    assert isinstance(captured.out, str)

    received = captured.out.strip().split("\n")
    assert len(received) == len(expected)

    for r, e in zip(received[1:], expected[1:]):  # FIXME: first row?
        assert r.find("\n") == -1
        assert r == e


def test_run_command_file_Excel(capsys):
    filename = (DATA_PATH / "actions.xlsx")
    assert filename.exists()

    expected = [
        f"Command file: {filename}",
        "====== ============ ===========================",
        "line # action       parameters                 ",
        "====== ============ ===========================",
        "1      mono_shutter open                       ",
        "2      USAXSscan    45.07, 98.3, 0, Water Blank",
        "3      saxsExp      45.07, 98.3, 0, Water Blank",
        "4      waxwsExp     45.07, 98.3, 0, Water Blank",
        "5      USAXSscan    12, 12, 1.2, plastic       ",
        "6      USAXSscan    12, 37, 0.1, Al foil       ",
        "7      mono_shutter close                      ",
        "====== ============ ===========================",
        "",
        "file line 1: ['mono_shutter', 'open', None, None, None]",
        "no handling for line 1: ['mono_shutter', 'open', None, None, None]",
        "file line 2: ['USAXSscan', 45.07, 98.3, 0, 'Water Blank']",
        "no handling for line 2: ['USAXSscan', 45.07, 98.3, 0, 'Water Blank']",
        "file line 3: ['saxsExp', 45.07, 98.3, 0, 'Water Blank']",
        "no handling for line 3: ['saxsExp', 45.07, 98.3, 0, 'Water Blank']",
        "file line 4: ['waxwsExp', 45.07, 98.3, 0, 'Water Blank']",
        "no handling for line 4: ['waxwsExp', 45.07, 98.3, 0, 'Water Blank']",
        "file line 5: ['USAXSscan', 12, 12, 1.2, 'plastic']",
        "no handling for line 5: ['USAXSscan', 12, 12, 1.2, 'plastic']",
        "file line 6: ['USAXSscan', 12, 37, 0.1, 'Al foil']",
        "no handling for line 6: ['USAXSscan', 12, 37, 0.1, 'Al foil']",
        "file line 7: ['mono_shutter', 'close', None, None, None]",
        "no handling for line 7: ['mono_shutter', 'close', None, None, None]",
    ]

    summarize_plan(run_command_file(filename))

    captured = capsys.readouterr()
    assert isinstance(captured.out, str)

    received = captured.out.strip().split("\n")
    assert len(received) == len(expected)

    for r, e in zip(received[1:], expected[1:]):  # FIXME: first row?
        assert r.find("\n") == -1
        assert r == e



"""
Ophyd support for the iocStats support


Public Structures

.. autosummary::

    ~IocStatsDevice

"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignalRO
from ophyd import Signal


class IocStatsDevice(Device):
    """
    synApps IOC stats

    .. index:: Ophyd Device; synApps IocStatsDevice
    """

    _app_dir1 = Component(EpicsSignalRO, "APP_DIR1", kind="omitted")
    _app_dir2 = Component(EpicsSignalRO, "APP_DIR2", kind="omitted")
    _startup_script1 = Component(EpicsSignalRO, "ST_SCRIPT1", kind="omitted")
    _startup_script2 = Component(EpicsSignalRO, "ST_SCRIPT2", kind="omitted")
    access = Component(EpicsSignalRO, "ACCESS", string=True, kind="config")
    ca_client_count = Component(EpicsSignalRO, "CA_CLNT_CNT")
    ca_connection_count = Component(EpicsSignalRO, "CA_CONN_CNT")
    cpu_count = Component(EpicsSignalRO, "CPU_CNT", kind="config")
    engineer = Component(EpicsSignalRO, "ENGINEER", kind="config")
    epics_version = Component(EpicsSignalRO, "EPICS_VERS", kind="config")
    file_descriptors_free = Component(EpicsSignalRO, "FD_FREE")
    file_descriptors_max_ = Component(EpicsSignalRO, "FD_MAX")
    heartbeat = Component(EpicsSignalRO, "HEARTBEAT", kind="omitted")
    host_name = Component(EpicsSignalRO, "HOSTNAME")
    ioc_cpu_load = Component(EpicsSignalRO, "IOC_CPU_LOAD")
    iso8601 = Component(EpicsSignalRO, "iso8601")
    kernel_version = Component(EpicsSignalRO, "KERNEL_VERS", kind="config")
    location = Component(EpicsSignalRO, "LOCATION")
    max_array_bytes = Component(EpicsSignalRO, "CA_MAX_ARRAY")
    memory_free = Component(EpicsSignalRO, "MEM_FREE")
    memory_max = Component(EpicsSignalRO, "MEM_MAX")
    memory_used = Component(EpicsSignalRO, "MEM_USED")
    records_count = Component(EpicsSignalRO, "RECORD_CNT")
    startup_time = Component(EpicsSignalRO, "STARTTOD", kind="config")
    suspended_task_count = Component(EpicsSignalRO, "SUSP_TASK_CNT")
    system_cpu_load = Component(EpicsSignalRO, "SYS_CPU_LOAD")
    time_of_day = Component(EpicsSignalRO, "TOD")
    timezone = Component(EpicsSignalRO, "TIMEZONE", kind="config")
    uptime = Component(EpicsSignalRO, ":UPTIME")

    # components updated from EpicsSignals
    application_directory = Component(Signal, value="", kind="config")
    memory_used_percentage = Component(Signal, value=0)
    startup_script = Component(Signal, value="")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def ad_update(value, *args, **kwargs):
            """
            update ``application_directory`` Signal

            Instead of a @property, keep a Signal updated
            so it will be recorded with a Device.read()
            """
            self.application_directory.put(self._app_dir1.get() + self._app_dir2.get())

        self._app_dir1.subscribe(ad_update)
        self._app_dir2.subscribe(ad_update)

        def ss_update(value, *args, **kwargs):
            """
            update ``startup_script`` Signal

            Instead of a @property, keep a Signal updated
            so it will be recorded with a Device.read()
            """
            self.startup_script.put(
                self._startup_script1.get() + self._startup_script2.get()
            )

        self._startup_script1.subscribe(ss_update)
        self._startup_script2.subscribe(ss_update)

        def mem_used_update(value, *args, **kwargs):
            """
            update ``memory_used_percentage`` Signal

            Instead of a @property, keep a Signal updated
            so it will be recorded with a Device.read()
            """
            self.memory_used_percentage.put(
                100 * self.memory_used.get() / self.memory_max.get()
            )

        self.memory_used.subscribe(mem_used_update)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Directory of bluesky runs
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~getRunData
   ~getRunDataValue
   ~listRunKeys
   ~ListRuns
   ~listruns
   ~summarize_runs
"""

import databroker
import databroker._drivers.mongo_normalized
import databroker._drivers.msgpack
import databroker.queries
import dataclasses
import datetime
import logging
import pandas as pd
import pyRestTable
import time
import typing

from collections import defaultdict

from .query import db_query

from ._core import FIRST_DATA
from ._core import LAST_DATA
from ._core import MONGO_CATALOG_CLASSES


logger = logging.getLogger(__name__)


def getRunData(scan_id, db=None, stream="primary", query=None, use_v1=True):
    """
    Convenience function to get the run's data.  Default is the ``primary`` stream.

    PARAMETERS

    scan_id
        *int* or *str* :
        Scan (run) identifier.
        Positive integer value is ``scan_id`` from run's metadata.
        Negative integer value is since most recent run in databroker.
        String is run's ``uid`` unique identifier (can abbreviate to
        the first characters needed to assure it is unique).

    db
        *object* :
        Bluesky database, an instance of ``databroker.catalog``.
        Default: will search existing session for instance.

    stream
        *str* :
        Name of the bluesky data stream to obtain the data.
        Default: 'primary'

    query
        *dict* :
        mongo query dictionary, used to filter the results
        Default: ``{}``

        see: https://docs.mongodb.com/manual/reference/operator/query/

    use_v1
        *bool* :
        Chooses databroker API version between 'v1' or 'v2'.
        Default: ``True`` (meaning use the v1 API)

    (new in apstools 1.5.1)
    """
    from . import getCatalog

    cat = getCatalog(db)
    if query:
        cat = db_query(cat, query)

    stream = stream or "primary"

    if use_v1 is None or use_v1:
        run = cat.v1[scan_id]
        if stream in run.stream_names:
            return run.table(stream_name=stream)
    else:
        run = cat.v2[scan_id]
        if hasattr(run, stream):
            return run[stream].read().to_dataframe()

    raise AttributeError(f"No such stream '{stream}' in run '{scan_id}'.")


def getRunDataValue(
    scan_id, key, db=None, stream="primary", query=None, idx=-1, use_v1=True
):
    """
    Convenience function to get value of key in run stream.

    Defaults are last value of key in primary stream.

    PARAMETERS

    scan_id
        *int* or *str* :
        Scan (run) identifier.
        Positive integer value is ``scan_id`` from run's metadata.
        Negative integer value is since most recent run in databroker.
        String is run's ``uid`` unique identifier (can abbreviate to
        the first characters needed to assure it is unique).

    key
        *str* :
        Name of the key (data column) in the table of the stream's data.
        Must match *identically*.

    db
        *object* :
        Bluesky database, an instance of ``databroker.catalog``.
        Default: will search existing session for instance.

    stream
        *str* :
        Name of the bluesky data stream to obtain the data.
        Default: 'primary'

    query
        *dict* :
        mongo query dictionary, used to filter the results
        Default: ``{}``

        see: https://docs.mongodb.com/manual/reference/operator/query/

    idx
        *int* or *str* :
        List index of value to be returned from column of table.
        Can be ``0`` for first value, ``-1`` for last value, ``"mean"``
        for average value, or ``"all"`` for the full list of values.
        Default: ``-1``

    use_v1
        *bool* :
        Chooses databroker API version between 'v1' or 'v2'.
        Default: ``True`` (meaning use the v1 API)

    (new in apstools 1.5.1)
    """
    if idx is None:
        idx = -1
    try:
        _idx = int(idx)
    except ValueError:
        _idx = str(idx).lower()

    if isinstance(_idx, str) and _idx not in "all mean".split():
        raise KeyError(
            f"Did not understand 'idx={idx}', use integer, 'all', or 'mean'."
        )

    stream = stream or "primary"

    table = getRunData(scan_id, db=db, stream=stream, query=query)

    if key not in table:
        raise KeyError(f"'{key}' not found in scan {scan_id} stream '{stream}'.")
    data = table[key]

    if _idx == "all":
        return data.values
    elif _idx == "mean":
        return data.mean()
    elif (0 <= _idx < len(data)) or (_idx < 0):
        return data.values[_idx]
    raise KeyError(
        f"Cannot reference idx={idx} in scan {scan_id} stream'{stream}' key={key}."
    )


def listRunKeys(
    scan_id,
    key_fragment="",
    db=None,
    stream="primary",
    query=None,
    strict=False,
    use_v1=True,
):
    """
    Convenience function to list all keys (column names) in the scan's stream (default: primary).

    PARAMETERS

    scan_id
        *int* or *str* :
        Scan (run) identifier.
        Positive integer value is ``scan_id`` from run's metadata.
        Negative integer value is since most recent run in databroker.
        String is run's ``uid`` unique identifier (can abbreviate to
        the first characters needed to assure it is unique).

    key_fragment
        *str* :
        Part or all of key name to be found in selected stream.
        For instance, if you specify ``key_fragment="lakeshore"``,
        it will return all the keys that include ``lakeshore``.

    db
        *object* :
        Bluesky database, an instance of ``databroker.catalog``.
        Default: will search existing session for instance.

    stream
        *str* :
        Name of the bluesky data stream to obtain the data.
        Default: 'primary'

    query
        *dict* :
        mongo query dictionary, used to filter the results
        Default: ``{}``

        see: https://docs.mongodb.com/manual/reference/operator/query/

    strict
        *bool* :
        Should the ``key_fragment`` be matched identically (``strict=True``)
        or matched by lower case comparison (``strict=False``)?
        Default: ``False``

    use_v1
        *bool* :
        Chooses databroker API version between 'v1' or 'v2'.
        Default: ``True`` (meaning use the v1 API)

    (new in apstools 1.5.1)
    """
    table = getRunData(scan_id, db=db, stream=stream, query=query, use_v1=use_v1)

    # fmt: off
    if len(key_fragment):
        output = [
            col
            for col in table.columns
            if (
                (strict and key_fragment in col)
                or (not strict and key_fragment.lower() in col.lower())
            )
        ]
    else:
        output = list(table.columns)
    # fmt: on
    return output


@dataclasses.dataclass
class ListRuns:
    """
    List the runs from the given catalog according to some options.

    EXAMPLE::

        ListRuns(cat).to_dataframe()

    PUBLIC METHODS

    .. autosummary::

        ~to_dataframe
        ~to_table
        ~parse_runs

    INTERNAL METHODS

    .. autosummary::

        ~_get_by_key
        ~_check_cat
        ~_apply_search_filters
        ~_check_keys

    """

    cat: object = None
    query: dict = None
    keys: str = None
    missing: str = ""
    num: int = 20
    reverse: bool = True
    since: str = None
    sortby: str = "time"
    timefmt: str = "%Y-%m-%d %H:%M:%S"
    until: str = None
    ids: "typing.Any" = None

    _default_keys = "scan_id time plan_name detectors"

    def _get_by_key(self, md, key):
        """
        Get run's metadata value by key.

        Look in ``start`` document first.
        If not found, look in ``stop`` document.
        If not found, report using ``self.missing``.

        If ``key`` is found but value is None, report as ``self.missing``.

        The ``time`` key will be formatted by the ``self.timefmt`` value.
        See https://strftime.org/ for examples.  The special ``timefmt="raw"``
        is used to report time as the raw value (floating point time as used in
        python's ``time.time()``).

        A special syntax of ``key`` allows reporting of keys in other
        metadata subdictionaries.  The syntax is ``doc.key`` (such as
        specifying the ``time`` key  from the ``stop`` document:
        ``stop.time``) where a ``.`` is used to separate the
        subdictionary name (``doc``) from the ``key``.
        **Note**:
        It is not possible to ``sortby`` the dotted-key syntax
        at this time.
        """
        v = None
        if key == "time":
            v = md["start"][key]
            if self.timefmt != "raw":
                ts = datetime.datetime.fromtimestamp(v)
                v = ts.strftime(self.timefmt)
        elif key in md["start"]:
            v = md["start"].get(key, self.missing)
        elif md["stop"] and key in md["stop"]:
            v = md["stop"].get(key, self.missing)
        elif len(key.split(".")) == 2:
            # dotted-key syntax
            doc, key = key.split(".")
            if md[doc] is not None:
                v = md[doc].get(key, self.missing)
                if key == "time" and self.timefmt != "raw":
                    ts = datetime.datetime.fromtimestamp(v)
                    v = ts.strftime(self.timefmt)
        return v or self.missing

    def _check_cat(self):
        from . import getCatalog

        if self.cat is None:
            self.cat = getCatalog()

    def _apply_search_filters(self):
        """Search for runs from the catalog."""
        since = self.since or FIRST_DATA
        until = self.until or LAST_DATA
        self._check_cat()
        query = {}
        query.update(databroker.queries.TimeRange(since=since, until=until))
        query.update(self.query or {})
        cat = self.cat.v2.search(query)
        return cat

    def parse_runs(self):
        """Parse the runs for the given metadata keys.  Return a dict."""
        self._check_keys()
        cat = self._apply_search_filters()

        def _sort(uid):
            """Sort runs in desired order based on metadata key."""
            md = self.cat[uid].metadata
            for doc in "start stop".split():
                if md[doc] and self.sortby in md[doc]:
                    return md[doc][self.sortby] or self.missing
            return self.missing

        num_runs_requested = min(abs(self.num), len(cat))
        results = {k: [] for k in self.keys}
        sequence = ()  # iterable of run uids

        if self.ids is not None:
            sequence = []
            for k in self.ids:
                try:
                    cat[k]  # try to access the run using `k`
                    sequence.append(k)
                except Exception as exc:
                    logger.warning(
                        "Could not find run %s in search of catalog %s: %s",
                        k,
                        self.cat.name,
                        exc,
                    )
        else:
            if isinstance(cat, MONGO_CATALOG_CLASSES) and self.sortby == "time":
                if self.reverse:
                    # the default rendering: from MongoDB in reverse time order
                    sequence = iter(cat)
                else:
                    # by increasing time order
                    sequence = [uid for uid in cat][::-1]
            else:
                # full search in Python
                sequence = sorted(cat.keys(), key=_sort, reverse=self.reverse)

        count = 0
        for uid in sequence:
            run = cat[uid]
            for k in self.keys:
                results[k].append(self._get_by_key(run.metadata, k))
            count += 1
            if count >= num_runs_requested:
                break
        return results

    def _check_keys(self):
        """Check that self.keys is a list of strings."""
        self.keys = self.keys or self._default_keys
        if isinstance(self.keys, str) and self.keys.find(" ") >= 0:
            # convert a space-delimited string of names
            self.keys = self.keys.split()

    def to_dataframe(self):
        """Output as pandas DataFrame object"""
        dd = self.parse_runs()
        return pd.DataFrame(dd, columns=self.keys)

    def to_table(self, fmt=None):
        """Output as pyRestTable object."""
        dd = self.parse_runs()

        table = pyRestTable.Table()
        rows = []
        for label, values in dd.items():
            table.addLabel(label)
            rows.append(values)
        table.rows = list(zip(*rows))

        return table.reST(fmt=fmt or "simple")


def listruns(
    cat=None,
    keys=None,
    missing="",
    num=20,
    printing="smart",
    reverse=True,
    since=None,
    sortby="time",
    tablefmt="dataframe",
    timefmt="%Y-%m-%d %H:%M:%S",
    until=None,
    ids=None,
    **query,
):
    """
    List runs from catalog.

    This function provides a thin interface to the highly-reconfigurable
    ``ListRuns()`` class in this package.

    PARAMETERS

    cat
        *object* :
        Instance of databroker v1 or v2 catalog.
    keys
        *str* or *[str]* or None:
        Include these additional keys from the start document.
        (default: ``None`` means ``"scan_id time plan_name detectors"``)
    missing
        *str*:
        Test to report when a value is not available.
        (default: ``""``)
    ids
        *[int]* or *[str]*:
        List of ``uid`` or ``scan_id`` value(s).
        Can mix different kinds in the same list.
        Also can specify offsets (e.g., ``-1``).
        According to the rules for ``databroker`` catalogs,
        a string is a ``uid`` (partial representations allowed),
        an int is ``scan_id`` if positive or an offset if negative.
        (default: ``None``)
    num
        *int* :
        Make the table include the ``num`` most recent runs.
        (default: ``20``)
    printing
        *bool* or ``"smart"``:
        If ``True``, print the table to stdout.
        If ``"smart"``, then act as shown below.
        (default: ``True``)

        ================  ===================
        session           action(s)
        ================  ===================
        python session    print and return ``None``
        Ipython console   return ``DataFrame`` object
        Jupyter notebook  return ``DataFrame`` object
        ================  ===================

    reverse
        *bool* :
        If ``True``, sort in descending order by ``sortby``.
        (default: ``True``)
    since
        *str* :
        include runs that started on or after this ISO8601 time
        (default: ``"1995-01-01"``)
    sortby
        *str* :
        Sort columns by this key, found by exact match in either
        the ``start`` or ``stop`` document.
        (default: ``"time"``)
    tablefmt
        *str* :
        When returning an object, specify which type
        of object to return.
        (default: ``"dataframe",``)

        ========== ==============
        value      object
        ========== ==============
        dataframe  ``pandas.DataFrame``
        table      ``str(pyRestTable.Table)``
        ========== ==============

    timefmt
        *str* :
        The ``time`` key (also includes keys ``"start.time"`` and  ``"stop.time"``)
        will be formatted by the ``self.timefmt`` value.
        See https://strftime.org/ for examples.  The special ``timefmt="raw"``
        is used to report time as the raw value (floating point time as used in
        python's ``time.time()``).
        (default: ``"%Y-%m-%d %H:%M:%S",``)
    until
        *str* :
        include runs that started before this ISO8601 time
        (default: ``2100-12-31``)
    ``**query``
        *dict* :
        Any additional keyword arguments will be passed to
        the databroker to refine the search for matching runs
        using the ``mongoquery`` package.

    RETURNS

    object:
        ``None`` or ``str`` or ``pd.DataFrame()`` object

    EXAMPLE::

        TODO

    (new in release 1.5.0)
    """

    lr = ListRuns(
        cat=cat,
        keys=keys,
        missing=missing,
        num=num,
        query=query,
        reverse=reverse,
        since=since,
        sortby=sortby,
        timefmt=timefmt,
        until=until,
        ids=ids,
    )

    tablefmt = tablefmt or "dataframe"
    if tablefmt == "dataframe":
        obj = lr.to_dataframe()
    else:
        obj = lr.to_table()

    if printing:
        if lr.cat is not None:
            print(f"catalog: {lr.cat.name}")
        print(obj)
        return
    return obj


def summarize_runs(since=None, db=None):
    """
    Report bluesky run metrics from the databroker.

    * How many different plans?
    * How many runs?
    * How many times each run was used?
    * How frequently?  (TODO:)

    PARAMETERS

    since
        *str* :
        Report all runs since this ISO8601 date & time
        (default: ``1995``)
    db
        *object* :
        Instance of ``databroker.Broker()``
        (default: ``db`` from the IPython shell)
    """
    from . import ipython_shell_namespace

    db = db or ipython_shell_namespace()["db"]
    # no APS X-ray experiment data before 1995!
    since = since or "1995"
    cat = db.v2.search(databroker.queries.TimeRange(since=since))
    plans = defaultdict(list)
    t0 = time.time()
    for n, uid in enumerate(cat):
        t1 = time.time()
        # next step is very slow (0.01 - 0.5 seconds each!)
        run = cat[uid]
        t2 = time.time()
        plan_name = run.metadata["start"].get("plan_name", "unknown")
        # fmt:off
        dt = datetime.datetime.fromtimestamp(
            run.metadata["start"]["time"]
        ).isoformat()
        # fmt:on
        scan_id = run.metadata["start"].get("scan_id", "unknown")
        # fmt: off
        plans[plan_name].append(
            dict(
                plan_name=plan_name,
                dt=dt,
                time_start=dt,
                uid=uid,
                scan_id=scan_id,
            )
        )
        # fmt: on
        logger.debug(
            "%s %s dt1=%4.01fus dt2=%5.01fms %s",
            scan_id,
            dt,
            (t1 - t0) * 1e6,
            (t2 - t1) * 1e3,
            plan_name,
        )
        t0 = time.time()

    def sorter(plan_name):
        return len(plans[plan_name])

    table = pyRestTable.Table()
    table.labels = "plan quantity".split()
    for k in sorted(plans.keys(), key=sorter, reverse=True):
        table.addRow((k, sorter(k)))
    table.addRow(("TOTAL", n + 1))
    print(table)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
DictionaryDevice
+++++++++++++++++++++++++++++++++++++++

Create an ophyd Device defined by a dictionary so
that a simple dictionary can be recorded in a bluesky data stream.

.. autosummary::

    ~dict_device_factory
    ~make_dict_device
"""

from ophyd import Component
from ophyd import Device
from ophyd import Signal
import time


def dict_device_factory(data={}):
    """
    Create a DictionaryDevice class using the supplied dictionary.
    """
    component_dict = {k: Component(Signal, value=v) for k, v in data.items()}
    fc = type("DictionaryDevice", (Device,), component_dict)
    return fc


def make_dict_device(obj, name="ddev"):
    """
    Make recordable DictionaryDevice instance from dictionary.

    New in release 1.6.4.
    """
    def standardize(obj):
        """Make obj look like .read()"""
        d_new = {}
        ts = time.time()  # new timestamps will be the same
        for k, v in obj.items():
            # fmt: off
            # Same structure as from Signal.read()?
            if (
                isinstance(v, dict) and
                sorted(list(v.keys())) == sorted("value timestamp".split())
            ):
                # looks like v is from .read(), accept it
                d_new[k] = v
            else:
                # Make it look like .read()
                d_new[k] = dict(value=v, timestamp=ts)
            # fmt: on
        return d_new

    obj = standardize(obj)
    kv = {k: v["value"] for k, v in obj.items()}
    ddev = dict_device_factory(kv)("", name=name)
    # set the timestamps to what was read
    for k, v in obj.items():
        getattr(ddev, k)._metadata["timestamp"] = v["timestamp"]
    return ddev


# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Miscellaneous Support
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~cleanupText
   ~connect_pvlist
   ~dictionary_table
   ~full_dotted_name
   ~itemizer
   ~listobjects
   ~pairwise
   ~print_RE_md
   ~redefine_motor_position
   ~replay
   ~run_in_thread
   ~safe_ophyd_name
   ~split_quoted_line
   ~text_encode
   ~to_unicode_or_bust
   ~trim_string_for_EPICS
   ~unix
"""

from ._core import MAX_EPICS_STRINGOUT_LENGTH
from ..callbacks import spec_file_writer
from .profile_support import ipython_shell_namespace
from bluesky import plan_stubs as bps
from bluesky.callbacks.best_effort import BestEffortCallback
from collections import OrderedDict
import databroker
import logging
import ophyd
import pyRestTable
import re
import subprocess
import sys
import threading
import time
import warnings


logger = logging.getLogger(__name__)


def cleanupText(text):
    """
    convert text so it can be used as a dictionary key

    Given some input text string, return a clean version
    remove troublesome characters, perhaps other cleanup as well.
    This is best done with regular expression pattern matching.
    """
    pattern = "[a-zA-Z0-9_]"

    def mapper(c):
        if re.match(pattern, c) is not None:
            return c
        return "_"

    return "".join([mapper(c) for c in text])


def dictionary_table(dictionary, **kwargs):
    """
    return a text table from ``dictionary``

    PARAMETERS

    dictionary
        *dict* :
        Python dictionary

    Note:  Keyword arguments parameters
    are kept for compatibility with previous
    versions of apstools.  They are ignored now.

    RETURNS

    table
        *object* or ``None`` :
        ``pyRestTable.Table()`` object (multiline text table)
        or ``None`` if dictionary has no contents

    EXAMPLE::

        In [8]: RE.md
        Out[8]: {'login_id': 'jemian:wow.aps.anl.gov', 'beamline_id': 'developer', 'proposal_id': None, 'pid': 19072, 'scan_id': 10, 'version': {'bluesky': '1.5.2', 'ophyd': '1.3.3', 'apstools': '1.1.5', 'epics': '3.3.3'}}

        In [9]: print(dictionary_table(RE.md))
        =========== =============================================================================
        key         value
        =========== =============================================================================
        beamline_id developer
        login_id    jemian:wow.aps.anl.gov
        pid         19072
        proposal_id None
        scan_id     10
        version     {'bluesky': '1.5.2', 'ophyd': '1.3.3', 'apstools': '1.1.5', 'epics': '3.3.3'}
        =========== =============================================================================

    """
    if len(dictionary) == 0:
        return
    t = pyRestTable.Table()
    t.addLabel("key")
    t.addLabel("value")
    for k, v in sorted(dictionary.items()):
        t.addRow((k, str(v)))
    return t


def full_dotted_name(obj):
    """
    Return the full dotted name

    The ``.dotted_name`` property does not include the
    name of the root object.  This routine adds that.

    see: https://github.com/bluesky/ophyd/pull/797
    """
    names = []
    while obj.parent is not None:
        names.append(obj.attr_name)
        obj = obj.parent
    names.append(obj.name)
    return ".".join(names[::-1])


def itemizer(fmt, items):
    """Format a list of items."""
    return [fmt % k for k in items]


def pairwise(iterable):
    """
    break a list (or other iterable) into pairs

    ::

        s -> (s0, s1), (s2, s3), (s4, s5), ...

        In [71]: for item in pairwise("a b c d e fg".split()):
            ...:     print(item)
            ...:
        ('a', 'b')
        ('c', 'd')
        ('e', 'fg')

    """
    a = iter(iterable)
    return zip(a, a)


def print_RE_md(dictionary=None, fmt="simple", printing=True):
    """
    custom print the RunEngine metadata in a table

    PARAMETERS

    dictionary
        *dict* :
        Python dictionary

    EXAMPLE::

        In [4]: print_RE_md()
        RunEngine metadata dictionary:
        ======================== ===================================
        key                      value
        ======================== ===================================
        EPICS_CA_MAX_ARRAY_BYTES 1280000
        EPICS_HOST_ARCH          linux-x86_64
        beamline_id              APS USAXS 9-ID-C
        login_id                 usaxs:usaxscontrol.xray.aps.anl.gov
        pid                      67933
        proposal_id              testing Bluesky installation
        scan_id                  0
        versions                 ======== =====
                                 key      value
                                 ======== =====
                                 apstools 1.1.3
                                 bluesky  1.5.2
                                 epics    3.3.1
                                 ophyd    1.3.3
                                 ======== =====
        ======================== ===================================

    """
    # override noting that fmt="markdown" will not display correctly
    fmt = "simple"

    dictionary = dictionary or ipython_shell_namespace()["RE"].md
    md = dict(dictionary)  # copy of input for editing
    v = dictionary_table(md["versions"])  # sub-table
    md["versions"] = v.reST(fmt=fmt).rstrip()
    table = dictionary_table(md)
    if printing:
        print("RunEngine metadata dictionary:")
        print(table.reST(fmt=fmt))
    return table


def replay(headers, callback=None, sort=True):
    """
    Replay the document stream from one (or more) scans (headers).

    PARAMETERS

    headers
        *run* or *[run]* :
        Run(s) to be replayed through callback. A *run* is an instance of a
        Bluesky ``databroker.core.BlueskyRun` (or the older
        ``databroker.Header``).
        see: https://nsls-ii.github.io/databroker/api.html?highlight=header#header-api

    callback
        *run* or *[run]* :
        The Bluesky callback to handle the stream of documents from a run. If
        ``None``, then use the `bec` (BestEffortCallback) from the IPython
        shell.
        (default:``None``)

    sort
        *bool* :
        Sort the headers chronologically if True.
        (default:``True``)

    (new in apstools release 1.1.11)
    """
    # fmt: off
    callback = callback or ipython_shell_namespace().get(
        "bec",  # get from IPython shell
        BestEffortCallback(),  # make one, if we must
    )
    # fmt: on
    _headers = headers  # do not mutate the input arg
    if not isinstance(_headers, (list, tuple)):
        _headers = [_headers]

    def as_header(run):
        if hasattr(run, "start"):
            return run  # this is a cat.v1 header
        # convert cat.v2 run to cat.v1 header
        uid = run.metadata["start"]["uid"]
        return run.cat.v1[uid]

    runs = [as_header(h) for h in _headers]

    def increasing_time_sorter(run):
        return run.start["time"]

    def decreasing_time_sorter(run):
        """Default for databroker v0 results."""
        return -run.start["time"]

    # fmt: off
    sorter = {
        True: increasing_time_sorter,
        False: decreasing_time_sorter,
    }[sort]
    # fmt: on

    for h in sorted(runs, key=sorter):
        if not isinstance(h, databroker.Header):
            # fmt: off
            raise TypeError(
                f"Must be a databroker Header: received: {type(h)}: |{h}|"
            )
            # fmt: on
        cmd = spec_file_writer._rebuild_scan_command(h.start)
        logger.debug("%s", cmd)

        # at last, this is where the real action happens
        for k, doc in h.documents():  # get the stream
            callback(k, doc)  # play it through the callback


def run_in_thread(func):
    """
    (decorator) run ``func`` in thread

    USAGE::

       @run_in_thread
       def progress_reporting():
           logger.debug("progress_reporting is starting")
           # ...

       #...
       progress_reporting()   # runs in separate thread
       #...

    """

    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return wrapper


def safe_ophyd_name(text):
    r"""
    make text safe to be used as an ophyd object name

    Given some input text string, return a clean version.
    Remove troublesome characters, perhaps other cleanup as well.
    This is best done with regular expression pattern matching.

    The "sanitized" name fits this regular expression::

        [A-Za-z_][\w_]*

    Also can be used for safe HDF5 and NeXus names.
    """
    replacement = "_"
    noncompliance = r"[^\w_]"

    # replace ALL non-compliances with '_'
    safer = replacement.join(re.split(noncompliance, text))

    # can't start with a digit
    if safer[0].isdigit():
        safer = replacement + safer
    return safer


def split_quoted_line(line):
    """
    splits a line into words some of which might be quoted

    TESTS::

        FlyScan 0   0   0   blank
        FlyScan 5   2   0   "empty container"
        FlyScan 5   12   0   "even longer name"
        SAXS 0 0 0 blank
        SAXS 0 0 0 "blank"

    RESULTS::

        ['FlyScan', '0', '0', '0', 'blank']
        ['FlyScan', '5', '2', '0', 'empty container']
        ['FlyScan', '5', '12', '0', 'even longer name']
        ['SAXS', '0', '0', '0', 'blank']
        ['SAXS', '0', '0', '0', 'blank']

    """
    parts = []

    # look for open and close quoted parts and combine them
    quoted = False
    multi = None
    for p in line.split():
        if not quoted and p.startswith('"'):  # begin quoted text
            quoted = True
            multi = ""

        if quoted:
            if len(multi) > 0:
                multi += " "
            multi += p
            if p.endswith('"'):  # end quoted text
                quoted = False

        if not quoted:
            if multi is not None:
                parts.append(multi[1:-1])  # remove enclosing quotes
                multi = None
            else:
                parts.append(p)

    return parts


def text_encode(source):
    """Encode ``source`` using the default codepoint."""
    return source.encode(errors="ignore")


def to_unicode_or_bust(obj, encoding="utf-8"):
    """from: http://farmdev.com/talks/unicode/  ."""
    if isinstance(obj, str):
        if not isinstance(obj, str):
            obj = str(obj, encoding)
    return obj


def trim_string_for_EPICS(msg):
    """String must not exceed EPICS PV length."""
    if len(msg) > MAX_EPICS_STRINGOUT_LENGTH:
        msg = msg[: MAX_EPICS_STRINGOUT_LENGTH - 1]
    return msg


def unix(command, raises=True):
    """
    Run a UNIX command, returns (stdout, stderr).

    PARAMETERS

    command
        *str* :
        UNIX command to be executed
    raises
        *bool* :
        If ``True``, will raise exceptions as needed,
        default: ``True``
    """
    if sys.platform not in ("linux", "linux2"):
        emsg = f"Cannot call unix() when OS={sys.platform}"
        raise RuntimeError(emsg)

    process = subprocess.Popen(
        command,
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    stdout, stderr = process.communicate()

    if len(stderr) > 0:
        emsg = f"unix({command}) returned error:\n{stderr}"
        logger.error(emsg)
        if raises:
            raise RuntimeError(emsg)

    return stdout, stderr


def listobjects(show_pv=True, printing=True, verbose=False, symbols=None):
    """
    Show all the ophyd Signal and Device objects defined as globals.

    PARAMETERS

    show_pv
        *bool* :
        If True, also show relevant EPICS PV, if available.
        (default: True)
    printing
        *bool* :
        If True, print table to stdout.
        (default: True)
    verbose
        *bool* :
        If True, also show ``str(obj``.
        (default: False)
    symbols
        *dict* :
        If None, use global symbol table.
        If not None, use provided dictionary.
        (default: ``globals()``)

    RETURNS

    object:
        Instance of ``pyRestTable.Table()``

    EXAMPLE::

        In [1]: listobjects()
        ======== ================================ =============
        name     ophyd structure                  EPICS PV
        ======== ================================ =============
        adsimdet MySingleTriggerSimDetector       vm7SIM1:
        m1       EpicsMotor                       vm7:m1
        m2       EpicsMotor                       vm7:m2
        m3       EpicsMotor                       vm7:m3
        m4       EpicsMotor                       vm7:m4
        m5       EpicsMotor                       vm7:m5
        m6       EpicsMotor                       vm7:m6
        m7       EpicsMotor                       vm7:m7
        m8       EpicsMotor                       vm7:m8
        noisy    EpicsSignalRO                    vm7:userCalc1
        scaler   ScalerCH                         vm7:scaler1
        shutter  SimulatedApsPssShutterWithStatus
        ======== ================================ =============

        Out[1]: <pyRestTable.rest_table.Table at 0x7fa4398c7cf8>

        In [2]:

    (new in apstools release 1.1.8)
    """
    table = pyRestTable.Table()
    table.labels = ["name", "ophyd structure"]
    if show_pv:
        table.addLabel("EPICS PV")
    if verbose:
        table.addLabel("object representation")
    table.addLabel("label(s)")
    if symbols is None:
        # the default choice
        g = ipython_shell_namespace()
        if len(g) == 0:
            # ultimate fallback
            g = globals()
    else:
        g = symbols
    for k, v in sorted(g.items()):
        if isinstance(v, (ophyd.Signal, ophyd.Device)):
            row = [k, v.__class__.__name__]
            if show_pv:
                if hasattr(v, "pvname"):
                    row.append(v.pvname)
                elif hasattr(v, "prefix"):
                    row.append(v.prefix)
                else:
                    row.append("")
            if verbose:
                row.append(str(v))
            row.append(" ".join(v._ophyd_labels_))
            table.addRow(row)
    if printing:
        print(table)
    return table


def connect_pvlist(pvlist, wait=True, timeout=2, poll_interval=0.1):
    """
    Given list of EPICS PV names, return dict of EpicsSignal objects.

    PARAMETERS

    pvlist
        *[str]* :
        list of EPICS PV names
    wait
        *bool* :
        should wait for EpicsSignal objects to connect
        (default: ``True``)
    timeout
        *float* :
        maximum time to wait for PV connections, seconds
        (default: 2.0)
    poll_interval
        *float* :
        time to sleep between checks for PV connections, seconds
        (default: 0.1)
    """
    obj_dict = OrderedDict()
    for item in pvlist:
        if len(item.strip()) == 0:
            continue
        pvname = item.strip()
        oname = "signal_{}".format(len(obj_dict))
        obj = ophyd.EpicsSignal(pvname, name=oname)
        obj_dict[oname] = obj

    if wait:
        times_up = time.time() + max(0, timeout)
        poll_interval = max(0.01, poll_interval)
        waiting = True
        while waiting and time.time() < times_up:
            time.sleep(poll_interval)
            waiting = False in [sig.connected for sig in obj_dict.values()]

        if waiting:
            # If did not connect all, revise with only the connected PVs
            # and report the unconncetd PVs.
            revised_dict = OrderedDict()
            for label, sig in obj_dict.items():
                if sig.connected:
                    revised_dict[label] = sig
                else:
                    print(f"Could not connect {sig.pvname}")
            if len(revised_dict) == 0:
                raise RuntimeError("Could not connect any PVs in the list")
            obj_dict = revised_dict

    return obj_dict


def redefine_motor_position(motor, new_position):
    """Set EPICS motor record's user coordinate to ``new_position``."""
    yield from bps.mv(motor.set_use_switch, 1)
    yield from bps.mv(motor.user_setpoint, new_position)
    yield from bps.mv(motor.set_use_switch, 0)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



# FIXME: refactor for new code

from .. import ActionsFlyerBase
from .. import FlyerBase
from .. import ScalerMotorFlyer
from .. import SignalValueStack
from ...tests import IOC
from bluesky import plans as bp
from bluesky import RunEngine
import databroker
import ophyd
import pysumreg
import pytest
import time
import uuid


cat = databroker.temp().v2
RE = RunEngine({})
RE.subscribe(cat.v1.insert)

m1 = ophyd.EpicsMotor(f"{IOC}m1", name="m1")
scaler1 = ophyd.scaler.ScalerCH(f"{IOC}scaler1", name="scaler1")
for obj in (m1, scaler1):
    obj.wait_for_connection()
scaler1.select_channels()

READING_KEY_LIST = []
for obj in (m1, scaler1):
    READING_KEY_LIST += list(obj.read().keys())
NUM_READING_KEYS = len(READING_KEY_LIST)
STREAM_NAME = "primary"
START = -1
FINISH = 1


def test_SignalValueStack():
    original = m1.motor_egu.get()
    new_text = str(uuid.uuid4())[:10]
    assert original != new_text

    stack = SignalValueStack()
    stack.remember(m1.motor_egu)
    assert m1.motor_egu.get() == original

    m1.motor_egu.put(new_text)
    assert m1.motor_egu.get(use_monitor=False) != original
    assert m1.motor_egu.get() == new_text

    stack.restore()
    assert m1.motor_egu.get(use_monitor=False) == original
    assert m1.motor_egu.get() != new_text


@pytest.mark.parametrize("flyer_class", [FlyerBase, ActionsFlyerBase])
def test_bases(flyer_class):
    flyer = flyer_class(name="flyer")
    assert isinstance(flyer, flyer_class)
    assert not hasattr(flyer, "action_exception")

    uids = RE(bp.fly([flyer]))
    assert len(uids) == 1

    run = cat.v2[uids[-1]]
    assert run is not None
    assert hasattr(run, "primary")

    dataset = run.primary.read()
    assert dataset is not None
    assert len(dataset.keys()) == 0

    assert not hasattr(flyer, "stats")


@pytest.mark.parametrize(
    "start, finish, fly_time, fly_time_pad, period, n_keys, n_data",
    [
        [-1, 1, None, None, None, NUM_READING_KEYS, 10],
        [-1, 1, 2, 2, 1, NUM_READING_KEYS, 2],
        [-1, 1, 1, 2, 0.1, NUM_READING_KEYS, 10],
        [0, 0, 1, 2, 0.1, NUM_READING_KEYS, 0],
        [1, -1, 1, 2, 0.1, NUM_READING_KEYS, 10],
    ],
)
def test_ScalerMotorFlyer(start, finish, fly_time, fly_time_pad, period, n_keys, n_data):
    flyer = ScalerMotorFlyer(
        scaler1, m1, start, finish, name="flyer", fly_time=fly_time, fly_time_pad=fly_time_pad, period=period
    )
    assert isinstance(flyer, ScalerMotorFlyer)
    assert hasattr(flyer, "action_exception")

    uids = RE(bp.fly([flyer]))
    assert flyer.action_exception is None
    assert len(uids) == 1

    run = cat.v2[uids[-1]]
    assert run is not None
    assert hasattr(run, "primary")

    dataset = run.primary.read()
    assert dataset is not None
    assert len(dataset.keys()) == n_keys

    key = READING_KEY_LIST[0]
    assert key in dataset

    array = dataset[key].values
    assert n_data <= len(array) <= n_data + 3

    # spot check some statistics
    assert hasattr(flyer, "stats")
    assert isinstance(flyer.stats, dict)

    skeys = list(flyer.stats.keys())
    assert len(skeys) < NUM_READING_KEYS
    for k in skeys:
        assert k in READING_KEY_LIST, k

    ch = f"{scaler1.time.name}"
    assert ch in flyer.stats

    ch_stats = flyer.stats[ch]
    assert isinstance(ch_stats, pysumreg.SummationRegisters)
    assert ch_stats.n == len(array)

    lo = min(start, finish)
    hi = max(start, finish)
    if ch_stats.n == 0:
        assert ch_stats.min_x is None
        assert ch_stats.max_x is None
        # avoid ZeroDivisionErrors
        stats_dict = ch_stats.to_dict()
        assert stats_dict["mean_x"] is None
        assert stats_dict["centroid"] is None
    else:
        assert ch_stats.min_x >= lo
        assert ch_stats.max_x <= hi
        assert lo <= ch_stats.mean_x <= hi
        assert lo <= ch_stats.centroid <= hi


@pytest.mark.parametrize(
    "scaler, motor, fly_time, fly_time_pad, period, ex_class, text",
    [
        [scaler1, m1, -1, None, None, ValueError, "POSITIVE number"],
        [scaler1, m1, "number", None, None, TypeError, ""],
        [scaler1, m1, 1, -1, None, ValueError, "POSITIVE number"],
        [scaler1, m1, 1, 1, -1, ValueError, "POSITIVE number"],
        [scaler1, scaler1, 1, 1, 1, TypeError, "Unprepared to handle as motor"],
        [m1, m1, 1, 1, 1, TypeError, "Unprepared to handle as scaler"],
    ],
)
def test_ScalerMotorFlyerExceptions(scaler, motor, fly_time, fly_time_pad, period, ex_class, text):
    with pytest.raises(ex_class) as exc:
        ScalerMotorFlyer(
            scaler, motor, -1, 1, name="flyer", fly_time=fly_time, fly_time_pad=fly_time_pad, period=period
        )
    assert text in str(exc.value)


@pytest.mark.parametrize(
    "period, ex_class, text",
    [
        [0.01, ValueError, "Could not get requested scaler sample period"],
    ],
)
def test_period_exception(period, ex_class, text):
    flyer = ScalerMotorFlyer(scaler1, m1, -1, 1, name="flyer", period=period)

    RE(bp.fly([flyer]))
    assert len(flyer._readings) == 0
    assert flyer.action_exception is not None
    assert isinstance(flyer.action_exception, ex_class)
    assert text in str(flyer.action_exception)



"""
Working with databroker catalogs
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~copy_filtered_catalog
   ~findCatalogsInNamespace
   ~getCatalog
   ~getDatabase
   ~getDefaultCatalog
   ~getDefaultDatabase
   ~getStreamValues
   ~quantify_md_key_use
"""

import databroker
import databroker.queries
import databroker._drivers.mongo_normalized
import databroker._drivers.msgpack
import logging
import pandas as pd
import pyRestTable


from .profile_support import getDefaultNamespace
from .list_runs import getRunData
from .profile_support import ipython_shell_namespace
from ._core import CATALOG_CLASSES


logger = logging.getLogger(__name__)


def copy_filtered_catalog(source_cat, target_cat, query=None):
    """
    copy filtered runs from source_cat to target_cat

    PARAMETERS

    source_cat
        *obj* :
        instance of `databroker.Broker` or `databroker.catalog[name]`
    target_cat
        *obj* :
        instance of `databroker.Broker` or `databroker.catalog[name]`
    query
        *dict* :
        mongo query dictionary, used to filter the results
        (default: ``{}``)

        see: https://docs.mongodb.com/manual/reference/operator/query/

    example::

        copy_filtered_catalog(
            databroker.Broker.named("mongodb_config"),
            databroker.catalog["test1"],
            {'plan_name': 'snapshot'})
    """
    query = query or {}
    for i, uid in enumerate(source_cat.v2.search(query)):
        run = source_cat.v1[uid]
        # fmt: off
        logger.debug(
            "%d  %s  #docs=%d", i + 1, uid, len(list(run.documents()))
        )
        # fmt: on
        for key, doc in run.documents():
            target_cat.v1.insert(key, doc)


def findCatalogsInNamespace():
    g = {}
    g.update(getDefaultNamespace())
    ns_cats = {}
    for k, v in g.items():
        if not k.startswith("_") and hasattr(v, "__class__"):
            try:
                if hasattr(v.v2, "container") and hasattr(v.v2, "metadata"):
                    ns_cats[k] = v
            except (AttributeError, TypeError):
                continue
    return ns_cats


def getCatalog(ref=None):
    if isinstance(ref, str):  # and ref in databroker.catalog:
        return databroker.catalog[ref]
    if ref is not None and hasattr(ref, "v2"):
        return ref.v2

    cat = getDefaultCatalog()
    if cat is None:
        raise ValueError("Cannot identify default databroker catalog.")
    return cat


def getDatabase(db=None, catalog_name=None):
    """
    Return Bluesky database using keyword guides or default choice.

    PARAMETERS

    db
        *object* :
        Bluesky database, an instance of ``databroker.catalog``
        (default: see ``catalog_name`` keyword argument)

    catalog_name
        *str* :
        Name of databroker v2 catalog, used when supplied
        ``db`` is ``None``.
        (default: catalog with most recent run timestamp)

    RETURNS

    object or ``None``:
        Bluesky database, an instance of ``databroker.catalog``

    (new in release 1.4.0)
    """
    if not hasattr(db, "v2"):
        # fmt: off
        if (
            hasattr(catalog_name, "name")
            and catalog_name in databroker.catalog
        ):
            # in case a catalog was passed as catalog_name
            db = catalog_name
        elif catalog_name is None:
            db = getDefaultDatabase()
        else:
            db = databroker.catalog[catalog_name]
        # fmt: on
    return db.v2


def getDefaultCatalog():
    cats = findCatalogsInNamespace()
    if len(cats) == 1:
        return cats[list(cats.keys())[0]]
    if len(cats) > 1:
        # fmt: off
        choices = "   ".join([
            f"{k} ({v.name})"
            for k, v in cats.items()
        ])
        # fmt: on
        raise ValueError(
            "No catalog defined.  "
            "Multiple catalog objects available."
            "  Specify one of these:"
            f" {choices}"
        )

    cats = list(databroker.catalog)
    if len(cats) == 1:
        return databroker.catalog[cats[0]]
    if len(cats) > 1:
        choices = "   ".join([f'databroker.catalog["{k}"]' for k in cats])
        raise ValueError(
            "No catalog defined.  "
            "Multiple catalog configurations available."
            "  Create specific catalog object from one of these commands:"
            f" {choices}"
        )

    raise ValueError("No catalogs available.")


def getDefaultDatabase():
    """
    Find the "default" database (has the most recent run).

    Note that here, *database* and *catalog* mean the same.

    This routine looks at all the database instances defined in the
    current session (console or notebook).  If there is only one or no
    database instances defined as objects in the current session, the
    choice is simple.  When there is more than one database instance in
    the current session, then the one with the most recent run timestamp
    is selected.  In the case (as happens when starting with a new
    database) that the current database has **no** runs *and* another
    database instance is defined in the session *and* that additional
    database has runs in it (such as the previous database), then the
    database with the newest run timestamp (and not the newer empty
    database) will be chosen.

    RETURNS

    object or ``None``:
        Bluesky database, an instance of ``databroker.catalog``

    (new in release 1.4.0)
    """
    # look through the console namespace
    g = ipython_shell_namespace()
    if len(g) == 0:
        # ultimate fallback
        g = globals()

    # note all database instances in memory
    db_list = []
    for v in g.values():
        if isinstance(v, CATALOG_CLASSES):
            db_list.append(v)

    # easy decisions first
    if len(db_list) == 0:
        return None
    if len(db_list) == 1:
        return db_list[0]

    # get the most recent run from each
    time_ref = {}
    for cat_name in list(databroker.catalog):
        cat = databroker.catalog[cat_name]
        if cat in db_list:
            if len(cat) > 0:
                run = cat.v2[-1]
                t = run.metadata["start"]["time"]
            else:
                t = 0
            time_ref[cat_name] = t, cat

    # pick the highest number for time
    highest = max([v[0] for v in time_ref.values()])
    choices = [v[1] for v in time_ref.values() if v[0] == highest]
    if len(choices) == 0:
        return None
    # return the catalog with the most recent timestamp
    return sorted(choices)[-1]


def getStreamValues(
    scan_id, key_fragment="", db=None, stream="baseline", query=None, use_v1=True
):
    """
    Get values from a previous scan stream in a databroker catalog.

    Optionally, select only those data with names including ``key_fragment``.

    .. tip::

        If the output is truncated, use
        ``pd.set_option('display.max_rows', 300)``
        to increase the number of rows displayed.

    PARAMETERS

    scan_id
        *int* or *str* :
        Scan (run) identifier.
        Positive integer value is ``scan_id`` from run's metadata.
        Negative integer value is since most recent run in databroker.
        String is run's ``uid`` unique identifier (can abbreviate to
        the first characters needed to assure it is unique).

    key_fragment
        *str* :
        Part or all of key name to be found in selected stream.
        For instance, if you specify ``key_fragment="lakeshore"``,
        it will return all the keys that include ``lakeshore``.

    db
        *object* :
        Bluesky database, an instance of ``databroker.catalog``.
        Default: will search existing session for instance.

    stream
        *str* :
        Name of the bluesky data stream to obtain the data.
        Default: 'baseline'

    query
        *dict* :
        mongo query dictionary, used to filter the results
        Default: ``{}``

        see: https://docs.mongodb.com/manual/reference/operator/query/

    use_v1
        *bool* :
        Chooses databroker API version between 'v1' or 'v2'.
        Default: ``True`` (meaning use the v1 API)

    RETURNS

    *object* :
        pandas DataFrame with values from selected stream, search_string, and query

        see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html

    (new in apstools 1.5.1)
    """
    if key_fragment is None:
        key_fragment = ""
    if use_v1 is None:
        use_v1 = True

    data = getRunData(scan_id, db=db, stream=stream, query=query, use_v1=use_v1)

    indices = [1, 2] if len(data["time"]) == 2 else [1]
    dd = {}

    # date_format = "%m/%d/%y %H:%M:%S"  # common in US
    date_format = "%y-%m-%d %H:%M:%S"  # modified ISO8601

    # fmt: off
    key = "time"
    dd[key] = [
        data[key][i].strftime(date_format)
        for i in indices
    ]
    # fmt: on

    for key in sorted(data.keys()):
        if key_fragment in key:
            dd[key] = [data[key][i] for i in indices]

    return pd.DataFrame(dd).transpose()


def quantify_md_key_use(
    key=None, db=None, catalog_name=None, since=None, until=None, query=None,
):
    """
    Print table of different ``key`` values and how many times each appears.

    PARAMETERS

    key *str* :
        one of the metadata keys in a run's start document
        (default: ``plan_name``)
    db *object* :
        Instance of databroker v1 ``Broker`` or v2 ``catalog``
        (default: see ``catalog_name`` keyword argument)
    catalog_name *str* :
        Name of databroker v2 catalog, used when supplied ``db`` is ``None``.
        (default: ``mongodb_config``)
    since *str* :
        include runs that started on or after this ISO8601 time
        (default: ``1995-01-01``)
    until *str* :
        include runs that started before this ISO8601 time
        (default: ``2100-12-31``)
    query *dict* :
        mongo query dictionary, used to filter the results
        (default: ``{}``)

        see: https://docs.mongodb.com/manual/reference/operator/query/

    EXAMPLES::

        quantify_md_key_use(key="proposal_id")
        quantify_md_key_use(key="plan_name", catalog_name="9idc", since="2020-07")
        quantify_md_key_use(key="beamline_id", catalog_name="9idc")
        quantify_md_key_use(key="beamline_id",
                            catalog_name="9idc",
                            query={'plan_name': 'Flyscan'},
                            since="2020",
                            until="2020-06-21 21:51")
        quantify_md_key_use(catalog_name="8id", since="2020-01", until="2020-03")

        In [8]: quantify_md_key_use(catalog_name="apstools_test")
        ========= =====
        plan_name #runs
        ========= =====
        count     26
        scan      27
        ========= =====

        In [9]: quantify_md_key_use(catalog_name="usaxs_test")
        ========================== =====
        plan_name                  #runs
        ========================== =====
        Flyscan                    1
        TuneAxis.tune              1
        count                      1
        measure_USAXS_Transmission 1
        run_Excel_file             1
        snapshot                   1
        tune_a2rp                  1
        tune_ar                    1
        tune_m2rp                  1
        tune_mr                    1
        ========================== =====
    """
    key = key or "plan_name"
    catalog_name = catalog_name or "mongodb_config"
    query = query or {}
    since = since or "1995-01-01"
    until = until or "2100-12-31"

    cat = (
        (db or databroker.catalog[catalog_name])
        .v2.search(databroker.queries.TimeRange(since=since, until=until))
        .search(query)
    )

    items = []
    while True:
        runs = cat.search({key: {"$exists": True, "$nin": items}})
        if len(runs) == 0:
            break
        else:
            items.append(runs.v1[-1].start.get(key))

    def sorter(key):
        if key is None:
            key = " None"
        return str(key)

    table = pyRestTable.Table()
    table.labels = f"{key} #runs".split()
    for item in sorted(items, key=sorter):
        table.addRow((item, len(cat.search({key: item}))))
    print(table)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Ophyd support for Stanford Research Systems 570 preamplifier from synApps

Public Structures

.. autosummary::

    ~SRS570_PreAmplifier

This device connects with the SRS570 support from synApps.
(https://github.com/epics-modules/ip/blob/master/ipApp/Db/SR570.db)

The SRS570 synApps support is part of the ``ip`` module:
https://htmlpreview.github.io/?https://raw.githubusercontent.com/epics-modules/ip/R3-6-1/documentation/swaitRecord.html

:see: https://github.com/epics-modules/ip
"""

from ophyd import Component
from ophyd import EpicsSignal
from .preamp_base import PreamplifierBaseDevice

import logging
import pint

logger = logging.getLogger(__name__)


class SRS570_PreAmplifier(PreamplifierBaseDevice):
    """
    Ophyd support for Stanford Research Systems 570 preamplifier from synApps.
    """

    # These values are strings because that is how they are defined
    # in the EPICS .db file.  Must cast them to ``float()`` or ``int()``
    # as desired.
    # see: https://github.com/epics-modules/ip/blob/master/ipApp/Db/SR570.db
    sensitivity_value = Component(EpicsSignal, "sens_num", kind="config", string=True)
    sensitivity_unit = Component(EpicsSignal, "sens_unit", kind="config", string=True)

    offset_on = Component(EpicsSignal, "offset_on", kind="config", string=True)
    offset_sign = Component(EpicsSignal, "offset_sign", kind="config", string=True)
    offset_value = Component(EpicsSignal, "offset_num", kind="config", string=True)
    offset_unit = Component(EpicsSignal, "offset_unit", kind="config", string=True)
    offset_fine = Component(EpicsSignal, "off_u_put", kind="config", string=True)
    offset_cal = Component(EpicsSignal, "offset_cal", kind="config", string=True)

    set_all = Component(EpicsSignal, "init.PROC", kind="config")

    bias_value = Component(EpicsSignal, "bias_put", kind="config", string=True)
    bias_on = Component(EpicsSignal, "bias_on", kind="config", string=True)

    filter_type = Component(EpicsSignal, "filter_type", kind="config", string=True)
    filter_lowpass = Component(EpicsSignal, "low_freq", kind="config", string=True)
    filter_highpass = Component(EpicsSignal, "high_freq", kind="config", string=True)

    gain_mode = Component(EpicsSignal, "gain_mode", kind="config", string=True)
    invert = Component(EpicsSignal, "invert_on", kind="config", string=True)
    blank = Component(EpicsSignal, "blank_on", kind="config", string=True)

    @property
    def computed_gain(self):
        """
        Amplifier gain (A/V), as floating-point number.
        """
        return pint.Quantity(
            float(self.sensitivity_value.get()),
            self.sensitivity_unit.get()
        ).to("A/V").magnitude

    def cb_gain(self, *args, **kwargs):
        """
        Called when sensitivity changes (EPICS CA monitor event).
        """
        self.gain.put(self.computed_gain)

    def __init__(self, *args, **kwargs):
        """
        Update the gain when the sensitivity changes.
        """
        super().__init__(*args, **kwargs)
        self.sensitivity_value.subscribe(self.cb_gain)
        self.sensitivity_unit.subscribe(self.cb_gain)

        # amplifier gain is the most relevant value
        self.gain.name = self.name

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Document Collector
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~DocumentCollectorCallback
   ~document_contents_callback
"""

import logging


logger = logging.getLogger(__name__).addHandler(logging.NullHandler())


def document_contents_callback(key, doc):
    """
    prints document contents -- use for diagnosing a document stream
    """
    print(key)
    for k, v in doc.items():
        print(f"\t{k}\t{v}")


class DocumentCollectorCallback(object):
    """
    Bluesky callback to collect *all* documents from most-recent plan

    .. index:: Bluesky Callback; DocumentCollectorCallback

    Will reset when it receives a *start* document.

    EXAMPLE::

        from apstools.callbacks import DocumentCollectorCallback
        doc_collector = DocumentCollectorCallback()
        RE.subscribe(doc_collector.receiver)
        ...
        RE(some_plan())
        print(doc_collector.uids)
        print(doc_collector.documents["stop"])

    """

    data_event_names = (
        "descriptor event resource datum bulk_events".split()
    )

    def __init__(self):
        self.documents = {}  # key: name, value: document
        self.uids = []  # chronological list of UIDs as-received

    def receiver(self, key, document):
        """keep all documents from recent plan in memory"""
        token = document.get("uid") or document.get("datum_id")
        if token is None:
            raise KeyError("No uid in '{}' document".format(key))
        self.uids.append(token)
        logger = logging.getLogger(__name__)
        logger.debug("%s document  uid=%s", key, str(token))  # lgtm [py/clear-text-logging-sensitive-data]
        if key == "start":
            self.documents = {key: document}
        elif key in self.data_event_names:
            if key not in self.documents:
                self.documents[key] = []
            self.documents[key].append(document)
        elif key == "stop":
            self.documents[key] = document
            print("exit status:", document["exit_status"])
            for item in self.data_event_names:
                if item in self.documents:
                    print(f"# {len(self.documents[item])}(s):")
        else:
            txt = "custom_callback encountered: %s\n%s"
            logger.warning(txt, key, document)
            if key not in self.documents:
                self.documents[key] = []
            self.documents[key].append(document)
        return

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Test execution of a blocking function in a plan.
"""

from ...tests import IOC
from ...utils import run_in_thread
from .. import run_blocking_function
from ophyd import EpicsSignal
import bluesky
import pytest
import random
import time

PV = f"{IOC}gp:float1"
T_MIN = 0.1
T_RANDOM_SCALE = 0.3
T_MAX = T_MIN + T_RANDOM_SCALE + 0.1


def create_signal():
    signal = EpicsSignal(PV, name="signal")
    signal.wait_for_connection()
    return signal


def procedure(signal, target_value):
    signal_now = signal.get(use_monitor=False)
    if signal_now == target_value:
        target_value += signal_now

    signal.put(target_value, use_complete=True)

    # finite, random delay, <= T_MAX
    time.sleep(T_MIN + T_RANDOM_SCALE * random.random())
    return target_value


@pytest.mark.parametrize(
    # fmt:off
    "target, expected, random_start",
    [
        [.1, .1, True],
        [.1, .2, False],
        [.2, .4, False],
        [.3, .3, False],
        [.1, .1, False],
    ]
    # fmt:on
)
def test_direct(target, expected, random_start):
    signal = create_signal()
    assert isinstance(signal, EpicsSignal)
    if random_start:
        signal.put(random.random())

    # test direct use, as from command-line, new target
    t0 = time.time()
    response = procedure(signal, target)
    dt = time.time() - t0
    assert T_MIN <= dt <= T_MAX
    assert response == expected
    assert signal.get() == expected


@pytest.mark.parametrize(
    # fmt:off
    "target, expected, random_start",
    [
        [.1, .1, True],
        [.1, .2, False],
        [.2, .4, False],
        [.3, .3, False],
        [.1, .1, False],
    ]
    # fmt:on
)
def test_bluesky(target, expected, random_start):
    signal = create_signal()
    assert isinstance(signal, EpicsSignal)
    if random_start:
        signal.put(random.random())

    RE = bluesky.RunEngine({})

    # test from RunEngine, different (original) target
    t0 = time.time()
    RE(run_blocking_function(procedure, signal, target))
    dt = time.time() - t0
    assert T_MIN <= dt <= T_MAX
    assert signal.get() == expected



"""
Issue 606

> While using TuneAxis().tune(), received this message from
> a tuning scan where fwhm=None. In this case, the scan did
> not cross the peak.

    Error saving stream PeakStats:

Because fwhm is None.
"""

from bluesky import RunEngine
import bluesky.plan_stubs as bps
from bluesky.callbacks import best_effort
import databroker
import dataclasses

from .. import TuneResults


@dataclasses.dataclass
class PeaksStatisticalData:
    """Statistical data from bec.peaks"""

    x: str = None
    y: str = None
    cen: float = None
    com: float = None
    fwhm: float = None
    min: list = None
    max: list = None
    crossings: list = None
    tune_ok: bool = False
    center: float = None
    initial_position: float = 0.0
    final_position: float = 0.0


def peak_stats_plan(peaks, stream="primary"):
    """Bluesky plan that saves PeakStats as a run."""
    stats = TuneResults("", name="stats")
    stats.set_stats(peaks)
    yield from bps.open_run(md={})
    yield from bps.create(name=stream)
    yield from bps.read(stats)
    yield from bps.save()
    yield from bps.close_run()


def test_results_stream(capsys):
    bec = best_effort.BestEffortCallback()
    cat = databroker.temp()
    RE = RunEngine({})
    RE.subscribe(cat.v1.insert)
    RE.subscribe(bec)

    peaks = PeaksStatisticalData()
    peaks.x = "m1"
    peaks.y = "noisy"
    peaks.cen = -0.1550618680443165
    peaks.com = -0.05085308669291895
    peaks.fwhm = None
    peaks.min = [6.65000000e-01, 7.08289488e02]
    peaks.max = [-3.35000000e-01, 6.30815141e03]
    peaks.crossings = [-0.15506187]
    peaks.tune_ok = True
    peaks.center = -0.1550618680443165
    peaks.initial_position = 0.165
    peaks.final_position = -0.1550618680443165

    RE(peak_stats_plan(peaks, "PeakStats"))
    out, err = capsys.readouterr()
    assert err == ""
    assert out.find("PeakStats") >= 0

    run = cat.v2[-1]
    assert "PeakStats" in run
    assert run.metadata["start"]["scan_id"] == 1

    ds = run.PeakStats.read()
    prefix = "stats_"
    assert ds[f"{prefix}x"].data[0] == "m1"
    assert ds[f"{prefix}y"].data[0] == "noisy"
    assert ds[f"{prefix}fwhm"].data[0] == "null"
    assert ds[f"{prefix}cen"].data[0] == -0.1550618680443165
    assert ds[f"{prefix}com"].data[0] == -0.05085308669291895
    assert ds[f"{prefix}crossings"].data[0] == [-0.15506187]

    # these keys are not written by TuneResults.set_stats()
    for k in "center tune_ok initial_position final_position".split():
        assert ds[f"{prefix}{k}"].data[0] == 0



import databroker._drivers.mongo_normalized
import databroker._drivers.msgpack
import intake


FIRST_DATA = "1995-01-01"
LAST_DATA = "2100-12-31"

CATALOG_CLASSES = (
    databroker.Broker,
    databroker._drivers.mongo_normalized.BlueskyMongoCatalog,
    databroker._drivers.msgpack.BlueskyMsgpackCatalog,
    intake.Catalog,
)
MONGO_CATALOG_CLASSES = (
    databroker.Broker,
    databroker._drivers.mongo_normalized.BlueskyMongoCatalog,
    # intake.Catalog,
)
MAX_EPICS_STRINGOUT_LENGTH = 40

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
